"""Summary
"""

import sys

import click

from ..cli.common import require_currency, require_environment, require_schema_type
from ..config import keyspace_types
from .schema import GraphsenseSchemas


def print_schema_validation_report(env, currency):
    report = GraphsenseSchemas().get_db_validation_report(env, currency)
    if any(report):
        for row in report:
            click.secho(f"Error: {row}", fg="red")
        sys.exit(101)
    else:
        click.echo(
            f"Db schema matches the expectation on {env} for currency {currency}."
        )


def keyspace_types_option(function):
    function = click.option(
        "--keyspace-type",
        type=click.Choice(keyspace_types, case_sensitive=False),
        help="Environment to work on",
        required=False,
    )(function)
    return function


@click.group()
def schema_cli():
    """ """
    pass


@schema_cli.group()
def schema():
    """Creating and validating the db schema."""
    pass


@schema.command(
    "show-by-currency",
    short_help="Prints the current db schema expected from graphsenselib",
)
@require_currency()
@keyspace_types_option
def showc(currency, keyspace_type):
    """Prints the current db schema expected from graphsenselib
    \f
    Args:
        currency (str): currency to work on
        keyspace_type (str): type of the keyspace
    """
    schemas = GraphsenseSchemas().get_by_currency(currency, keyspace_type=keyspace_type)
    for file, schema in schemas:
        click.echo(f"// ######### {file}")
        click.echo(schema.original_schema)


@schema.command(
    "show-by-schema-type",
    short_help="Prints the current db schema expected from graphsenselib",
)
@require_schema_type()
@keyspace_types_option
def shows(schema, keyspace_type):
    """Prints the current db schema expected from graphsenselib
    \f
    Args:
        currency (str): currency to work on
        keyspace_type (str): type of the keyspace
    """
    schemas = GraphsenseSchemas().get_by_schema_type(
        schema, keyspace_type=keyspace_type
    )
    for file, schema in schemas:
        click.echo(f"// ######### {file}")
        click.echo(schema.original_schema)


@schema.command(
    "create", short_help="Creates the necessary graphsense tables in Cassandra."
)
@require_environment()
@require_currency()
@keyspace_types_option
def create(env, currency, keyspace_type):
    """Summary
        Creates the necessary graphsense tables in Cassandra if they don't exist.
        \f
    Args:
        env (str): Environment to work on
        currency (str): currency to work on
    """
    if keyspace_type is None:
        GraphsenseSchemas().create_keyspaces_if_not_exist(env, currency)
    else:
        GraphsenseSchemas().create_keyspace_if_not_exist(
            env, currency, keyspace_type=keyspace_type
        )

    click.echo("// ######### Validating deployed schema")
    print_schema_validation_report(env, currency)


@schema.command(
    "create-new-transformed", short_help="Creates new/empty transformed keyspace."
)
@require_environment()
@require_currency()
@click.option(
    "--suffix",
    type=str,
    required=False,
    help="suffix to append to default "
    "[currency]_transformed_[date]_[suffix] keyspace name.",
)
@click.option(
    "--suffix",
    type=str,
    required=False,
    help="suffix to append to default "
    "[currency]_transformed_[date]_[suffix] keyspace name.",
)
@click.option(
    "--no-date",
    is_flag=True,
    help="omits the date in the keyspace name",
)
def create_new_tf(env, currency, suffix, no_date):
    """Summary
        Creates new/empty transformed keyspace.
        \f
    Args:
        env (str): Environment to work on
        currency (str): currency to work on
        suffix (str): suffix for the new keyspace
    """
    print(
        GraphsenseSchemas().create_new_transformed_ks_if_not_exist(
            env, currency, suffix, no_date
        )
    )


@schema.command(
    "validate", short_help="Validates if the expected schema matches the database."
)
@require_environment()
@require_currency()
def validate(env, currency):
    """Summary
        Validates if the expected schema matches the database.
    \f
    Args:
        env (str): Environment to work on
        currency (str): currency to work on
    """
    print_schema_validation_report(env, currency)



from collections import Counter
from dataclasses import dataclass
from typing import Iterable, List, Set

from ..datatypes import FlowDirection
from .generic import flatten


@dataclass
class SlimTx:
    address: str
    block_id: int
    timestamp: int
    tx_hash: str
    direction: FlowDirection


def filter_inoutputs(inoutputs_list: list) -> List:
    if inoutputs_list is None:
        return []
    return [
        (inout.address[0], inout.value)
        for inout in inoutputs_list
        if inout.address is not None and len(inout.address) == 1
    ]


def regularize_inoutputs(inoutputs_list: list) -> dict:
    """
    Sums the in/outputs for the same address in a inputput list of a tx

    Args:
        inoutputs_list (list): Inout-list like stored in raw keyspace

    Returns:
        dict: keyed by address.
    """
    # Ignore inouts with more than one address (copied form spark transform)
    if inoutputs_list is None:
        return {}
    r = filter_inoutputs(inoutputs_list)
    cnt = Counter()
    for adr, value in r:
        cnt[adr] += value
    return dict(cnt)


def get_total_input_sum(input_list: list) -> int:
    """Simple sum of all input

    Args:
        input_list (list): list of inputs

    Returns:
        int: sum of input
    """
    if input_list is None:
        return 0
    return sum([inp.value for inp in input_list])


def get_regflow(regin: dict, regout: dict, address: str) -> int:
    """Calculates the in/out flow an address.

    Args:
        regin (dict): regularized inputs
        regout (dict): regularized outputs
        address (str): address of interest

    Returns:
        int: Negative if net outflow, positive if net inflow
    """
    return regout.get(address, 0) - regin.get(address, 0)


def get_unique_addresses_from_transaction(transaction) -> Iterable[str]:
    return {adr.address for adr in get_slim_tx_from_transaction(transaction)}


def get_slim_tx_from_transaction(transaction) -> Iterable[SlimTx]:
    # Only take first address from address array
    # this is equivalent to the spark job, but is ignoring multisig

    output_addresses = [
        (address, FlowDirection.OUT)
        for (address, value) in filter_inoutputs(transaction.outputs)
    ]
    input_addresses = [
        (address, FlowDirection.IN)
        for (address, value) in filter_inoutputs(transaction.inputs)
    ]

    addresses = input_addresses + output_addresses
    return [
        SlimTx(
            addr,
            transaction.block_id,
            transaction.timestamp,
            transaction.tx_hash,
            direction,
        )
        for addr, direction in addresses
    ]


def get_slim_tx_from_transactions(transactions) -> Iterable[SlimTx]:
    return flatten([get_slim_tx_from_transaction(tx) for tx in transactions])


def get_unique_addresses_from_transactions(transactions) -> Set[str]:
    return {adr.address for adr in get_slim_tx_from_transactions(transactions)}


def get_unique_ordered_output_addresses_from_transactions(
    transactions,
) -> Iterable[str]:
    """Returns all unique output addresses in the order they appear in the txs.
    This is useful to assign address ids where order should matter.

    Args:
        transactions (TYPE): Iterable of dbtransactions

    Returns:
        Iterable[str]: order preserving Iterable
    """
    """
        Construction see
        https://stackoverflow.com/questions/1653970/does-python-have-an-ordered-set
    """
    return list(
        dict.fromkeys(
            [
                tx.address
                for tx in get_slim_tx_from_transactions(transactions)
                if tx.direction == FlowDirection.OUT
            ]
        )
    )


def get_unique_ordered_input_addresses_from_transactions(
    transactions,
) -> Iterable[str]:
    """Returns all unique input addresses in the order they appear in the txs.
    This is useful to assign address ids where order should matter.

    Args:
        transactions (TYPE): Iterable of dbtransactions

    Returns:
        Iterable[str]: order preserving Iterable
    """
    """
        Construction see
        https://stackoverflow.com/questions/1653970/does-python-have-an-ordered-set
    """
    return list(
        dict.fromkeys(
            [
                tx.address
                for tx in get_slim_tx_from_transactions(transactions)
                if tx.direction == FlowDirection.IN
            ]
        )
    )



from enum import Enum


class EntityType(Enum):
    CLUSTER = "cluster"
    ADDRESS = "address"

    def __str__(self):
        return str(self.value)


class DbChangeType(Enum):
    NEW = "new"
    UPDATE = "update"
    TRUNCATE = "truncate"
    DELETE = "delete"

    def __str__(self):
        return str(self.value)


class FlowDirection(Enum):
    IN = "in"
    OUT = "out"

    def __str__(self):
        return str(self.value)


# TODO add as enums and remove from config
# schema_types = ["utxo", "account"]
# keyspace_types = ["raw", "transformed"]



from graphsenselib.deltaupdate.update.generic import DbDelta
from graphsenselib.deltaupdate.update.utxo.update import dbdelta_from_utxo_transaction
from graphsenselib.utils import group_by, groupby_property
from graphsenselib.utils.utxo import get_unique_addresses_from_transaction

from .test_data import get_arel, get_atxs, get_exchange_rates_per_block, get_txs


def test_transaction_changeset_address_txs():
    rates = get_exchange_rates_per_block()
    expected_atxs = get_atxs()

    expected = groupby_property(expected_atxs, "tx_id", sort_by="identifier")
    atxs = []
    for tx in get_txs():
        cngset = dbdelta_from_utxo_transaction(tx, rates[tx.block_id])
        atxs.extend(list(cngset.new_entity_txs))
    res = groupby_property(atxs, "tx_id", sort_by="identifier")

    assert res == expected


def test_merge_address_deltas():
    rates = get_exchange_rates_per_block()
    changes = DbDelta.merge(
        [dbdelta_from_utxo_transaction(tx, rates[tx.block_id]) for tx in get_txs()]
    )

    expected = get_arel()

    expected_grouped = group_by(expected, key=lambda x: (x.src_address, x.dst_address))

    changes_grouped = group_by(
        changes.relation_updates, key=lambda x: (x.src_identifier, x.dst_identifier)
    )

    assert len(expected) == len(changes.relation_updates)

    for k, v in changes_grouped.items():
        ev = expected_grouped[k]
        assert len(v) == 1
        assert len(ev) == 1
        v = v[0]
        ev = ev[0]
        assert ev.estimated_value == v.estimated_value


def test_address_updates_order():
    rates = get_exchange_rates_per_block()
    changes = DbDelta.merge(
        [dbdelta_from_utxo_transaction(tx, rates[tx.block_id]) for tx in get_txs()]
    )
    last_tx_id = (-1, -1)
    for cng in changes.entity_updates:
        assert last_tx_id <= (cng.first_tx_id, cng.last_tx_id)
        last_tx_id = (cng.first_tx_id, cng.last_tx_id)


def test_address_addresses_unique():
    rates = get_exchange_rates_per_block()
    changes = DbDelta.merge(
        [dbdelta_from_utxo_transaction(tx, rates[tx.block_id]) for tx in get_txs()]
    )
    addr = [chng.identifier for chng in changes.entity_updates]
    assert len(addr) == len(set(addr))


def test_unique_addresses_in_tx():
    assert get_unique_addresses_from_transaction(get_txs()[5]) == {
        "3Fkx2TFdcHoab4xGgSjhAVh5YBPvbBWjNL",
        "1FAkhqm95YnV5Mi7Q5j2Wb8CkbK7Z9zpyB",
    }



import pytest
from diskcache import Cache

from graphsenselib.utils.cache import TableBasedCache


def test_cache():
    c = TableBasedCache(Cache())

    c[("abc", 1)] = 1
    c[("abc", 2)] = 2

    assert c[("abc", 1)] == 1
    assert c[("abc", 2)] == 2

    c.put_items("abc", [(1, 3), (2, 4)])

    assert c[("abc", 1)] == 3
    assert c[("abc", 2)] == 4

    data = [
        {"a": 1, "b": 2},
        {"a": 3, "b": 2},
        {"a": 1, "b": 3},
    ]

    c.put_items_keyed_by("abc", data, key="a")

    assert len(c.get_item("abc", 1)) == 2
    assert c.get_item("abc", 1) == [{"a": 1, "b": 2}, {"a": 1, "b": 3}]
    assert len(c.get_item("abc", 3)) == 1
    assert c.get_item("abc", 2) == 4

    c.delete_item("abc", 1)
    with pytest.raises(KeyError):
        assert len(c.get_item("abc", 1)) == 2

    assert c.get_item("abc", 2) == 4

    del c[("abc", 2)]

    with pytest.raises(KeyError):
        c.get_item("abc", 2)

    assert c.get(("abc", 2), None) is None
    assert c.get(("abc", 2), []) == []



# flake8: noqa

import logging
import unittest
from collections import defaultdict
from datetime import datetime
from typing import Any, Callable, Dict, List, NamedTuple, Tuple
from unittest.mock import MagicMock, patch

import graphsenselib
from graphsenselib.db import DbChange
from graphsenselib.deltaupdate.update.abstractupdater import TABLE_NAME_DELTA_HISTORY
from graphsenselib.deltaupdate.update.account.createchanges import (
    prepare_balances_for_ingest,
    prepare_entities_for_ingest,
    prepare_entity_txs_for_ingest,
    prepare_relations_for_ingest,
    prepare_txs_for_ingest,
)
from graphsenselib.deltaupdate.update.account.modelsdelta import (
    BalanceDelta,
    EntityDeltaAccount,
    RawEntityTxAccount,
    RelationDeltaAccount,
)
from graphsenselib.deltaupdate.update.generic import DeltaScalar, Tx
from graphsenselib.utils import DataObject as MutableNamedTuple
from graphsenselib.utils.account import (
    get_id_group,
    get_id_group_with_secondary_addresstransactions,
    get_id_group_with_secondary_relations,
)
from graphsenselib.utils.logging import LoggerScope


class TestPrepareTxsForIngest(unittest.TestCase):
    def setUp(self):
        self.mock_delta = [
            Tx(tx_id=1, tx_hash=b"0x1234567", block_id=123, failed=False, tx_index=0),
            Tx(
                tx_id=2, tx_hash=b"0x1222234", block_id=234, failed=True, tx_index=1
            ),  # This should be skipped in block_transactions
            # Add more mock Tx objects as needed
        ]
        self.id_bucket_size = 10
        self.block_bucket_size = 5
        self.mock_get_transaction_prefix = MagicMock(return_value=("prefix", "1"))

    def test_prepare_txs_for_ingest(self):
        # Mock the external function behaviors
        self.mock_get_transaction_prefix.return_value = ("prefix", "1")

        changes = prepare_txs_for_ingest(
            self.mock_delta,
            self.id_bucket_size,
            self.block_bucket_size,
            self.mock_get_transaction_prefix,
        )
        print(changes)

        n_changes_transaction_ids_by_transaction_id_group = len(
            [x for x in changes if x.table == "transaction_ids_by_transaction_id_group"]
        )
        n_changes_block_transactions = len(
            [x for x in changes if x.table == "block_transactions"]
        )

        self.assertEqual(n_changes_transaction_ids_by_transaction_id_group, 2)
        self.assertEqual(n_changes_block_transactions, 1)


class TestPrepareBalancesForIngest(unittest.TestCase):
    def setUp(self):
        # Mock BalanceDelta objects and addr_balances
        self.mock_delta = [
            BalanceDelta(identifier=123, asset_balances={"USDT": DeltaScalar(1)}),
            BalanceDelta(
                identifier=234,
                asset_balances={"USDT": DeltaScalar(2), "ETH": DeltaScalar(1)},
            ),
            # BalanceDelta(identifier=123, asset_balances={'USDT': DeltaScalar(2)}), cant be twice because it is compressed before
            BalanceDelta(identifier=456, asset_balances={"ETH": DeltaScalar(2)}),
        ]
        self.addr_balances = {
            234: BalanceDelta(identifier=234, asset_balances={"USDT": DeltaScalar(2)}),
            123: BalanceDelta(
                identifier=123,
                asset_balances={"USDT": DeltaScalar(2), "ETH": DeltaScalar(1)},
            ),
        }

        self.addr_balances_ref = {
            234: {"USDT": 4, "ETH": 1},
            123: {"USDT": 3},
            456: {"ETH": 2},
        }

        self.id_bucket_size = 100

    def test_prepare_balances_for_ingest(self):
        changes = prepare_balances_for_ingest(
            self.mock_delta, self.id_bucket_size, self.addr_balances
        )

        print(changes)
        # Check the total number of DbChange instances created
        self.assertEqual(
            len(changes), 4, "Incorrect number of DbChange instances created"
        )

        # transform DbChange to dict and compare with expected
        change_dict = defaultdict(dict)

        for change in changes:
            data = change.data
            change_dict[data["address_id"]][data["currency"]] = data["balance"]

        for k, v in self.addr_balances_ref.items():
            for currency, balance in v.items():
                self.assertEqual(change_dict[k][currency], balance)


class TestPrepareEntityTxsForIngest(unittest.TestCase):
    def setUp(self):
        # Setup your mock data here
        self.mock_delta = [
            RawEntityTxAccount(
                identifier=222,
                tx_id=123,
                block_id=100,
                is_outgoing=True,
                tx_reference="ref1",  # should be UserType but its fine
                token_values={"tokenA": 100, "tokenB": 200},
                value=0,  #  A RawEntityTxAccount can only have either value or token_values the way the functions are written. Could write a unit test for that too
            ),
            RawEntityTxAccount(
                identifier=111,
                tx_id=234,
                block_id=200,
                is_outgoing=False,
                tx_reference="ref2",
                token_values={},  # This should test the non-token transfer scenario
                value=1,
            ),
        ]
        self.id_bucket_size = 10
        self.currency = "currency1"
        self.block_bucket_size_address_txs = 5

        self.expected_changes_count = 3

    def test_prepare_entity_txs_for_ingest(self):
        changes = prepare_entity_txs_for_ingest(
            self.mock_delta,
            self.id_bucket_size,
            self.currency,
            self.block_bucket_size_address_txs,
        )

        print(changes)

        self.assertEqual(len(changes), self.expected_changes_count)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graphsenselib/ingest/tron/grpc/core/chain.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from graphsenselib.ingest.tron.grpc.core import common_pb2 as graphsenselib_dot_ingest_dot_tron_dot_grpc_dot_core_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/graphsenselib/ingest/tron/grpc/core/chain.proto\x12\x08protocol\x1a\x19google/protobuf/any.proto\x1a\x30graphsenselib/ingest/tron/grpc/core/common.proto\"\xc4\x11\n\x0bTransaction\x12+\n\x08raw_data\x18\x01 \x01(\x0b\x32\x19.protocol.Transaction.raw\x12\x11\n\tsignature\x18\x02 \x03(\x0c\x12)\n\x03ret\x18\x05 \x03(\x0b\x32\x1c.protocol.Transaction.Result\x1a\xf9\x08\n\x08\x43ontract\x12\x39\n\x04type\x18\x01 \x01(\x0e\x32+.protocol.Transaction.Contract.ContractType\x12\'\n\tparameter\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\x12\x10\n\x08provider\x18\x03 \x01(\x0c\x12\x14\n\x0c\x43ontractName\x18\x04 \x01(\x0c\x12\x15\n\rPermission_id\x18\x05 \x01(\x05\"\xc9\x07\n\x0c\x43ontractType\x12\x19\n\x15\x41\x63\x63ountCreateContract\x10\x00\x12\x14\n\x10TransferContract\x10\x01\x12\x19\n\x15TransferAssetContract\x10\x02\x12\x15\n\x11VoteAssetContract\x10\x03\x12\x17\n\x13VoteWitnessContract\x10\x04\x12\x19\n\x15WitnessCreateContract\x10\x05\x12\x16\n\x12\x41ssetIssueContract\x10\x06\x12\x19\n\x15WitnessUpdateContract\x10\x08\x12!\n\x1dParticipateAssetIssueContract\x10\t\x12\x19\n\x15\x41\x63\x63ountUpdateContract\x10\n\x12\x19\n\x15\x46reezeBalanceContract\x10\x0b\x12\x1b\n\x17UnfreezeBalanceContract\x10\x0c\x12\x1b\n\x17WithdrawBalanceContract\x10\r\x12\x19\n\x15UnfreezeAssetContract\x10\x0e\x12\x17\n\x13UpdateAssetContract\x10\x0f\x12\x1a\n\x16ProposalCreateContract\x10\x10\x12\x1b\n\x17ProposalApproveContract\x10\x11\x12\x1a\n\x16ProposalDeleteContract\x10\x12\x12\x18\n\x14SetAccountIdContract\x10\x13\x12\x12\n\x0e\x43ustomContract\x10\x14\x12\x17\n\x13\x43reateSmartContract\x10\x1e\x12\x18\n\x14TriggerSmartContract\x10\x1f\x12\x0f\n\x0bGetContract\x10 \x12\x19\n\x15UpdateSettingContract\x10!\x12\x1a\n\x16\x45xchangeCreateContract\x10)\x12\x1a\n\x16\x45xchangeInjectContract\x10*\x12\x1c\n\x18\x45xchangeWithdrawContract\x10+\x12\x1f\n\x1b\x45xchangeTransactionContract\x10,\x12\x1d\n\x19UpdateEnergyLimitContract\x10-\x12#\n\x1f\x41\x63\x63ountPermissionUpdateContract\x10.\x12\x14\n\x10\x43learABIContract\x10\x30\x12\x1b\n\x17UpdateBrokerageContract\x10\x31\x12\x1c\n\x18ShieldedTransferContract\x10\x33\x12\x1b\n\x17MarketSellAssetContract\x10\x34\x12\x1d\n\x19MarketCancelOrderContract\x10\x35\x1a\xce\x05\n\x06Result\x12\x0b\n\x03\x66\x65\x65\x18\x01 \x01(\x03\x12.\n\x03ret\x18\x02 \x01(\x0e\x32!.protocol.Transaction.Result.code\x12@\n\x0b\x63ontractRet\x18\x03 \x01(\x0e\x32+.protocol.Transaction.Result.contractResult\x12\x14\n\x0c\x61ssetIssueID\x18\x0e \x01(\t\x12\x17\n\x0fwithdraw_amount\x18\x0f \x01(\x03\x12\x17\n\x0funfreeze_amount\x18\x10 \x01(\x03\x12 \n\x18\x65xchange_received_amount\x18\x12 \x01(\x03\x12&\n\x1e\x65xchange_inject_another_amount\x18\x13 \x01(\x03\x12(\n exchange_withdraw_another_amount\x18\x14 \x01(\x03\x12\x13\n\x0b\x65xchange_id\x18\x15 \x01(\x03\x12 \n\x18shielded_transaction_fee\x18\x16 \x01(\x03\"\x1e\n\x04\x63ode\x12\n\n\x06SUCESS\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\"\xb1\x02\n\x0e\x63ontractResult\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x0b\n\x07SUCCESS\x10\x01\x12\n\n\x06REVERT\x10\x02\x12\x18\n\x14\x42\x41\x44_JUMP_DESTINATION\x10\x03\x12\x11\n\rOUT_OF_MEMORY\x10\x04\x12\x18\n\x14PRECOMPILED_CONTRACT\x10\x05\x12\x13\n\x0fSTACK_TOO_SMALL\x10\x06\x12\x13\n\x0fSTACK_TOO_LARGE\x10\x07\x12\x15\n\x11ILLEGAL_OPERATION\x10\x08\x12\x12\n\x0eSTACK_OVERFLOW\x10\t\x12\x11\n\rOUT_OF_ENERGY\x10\n\x12\x0f\n\x0bOUT_OF_TIME\x10\x0b\x12\x17\n\x13JVM_STACK_OVER_FLOW\x10\x0c\x12\x0b\n\x07UNKNOWN\x10\r\x12\x13\n\x0fTRANSFER_FAILED\x10\x0e\x1a\xfc\x01\n\x03raw\x12\x17\n\x0fref_block_bytes\x18\x01 \x01(\x0c\x12\x15\n\rref_block_num\x18\x03 \x01(\x03\x12\x16\n\x0eref_block_hash\x18\x04 \x01(\x0c\x12\x12\n\nexpiration\x18\x08 \x01(\x03\x12\"\n\x05\x61uths\x18\t \x03(\x0b\x32\x13.protocol.authority\x12\x0c\n\x04\x64\x61ta\x18\n \x01(\x0c\x12\x30\n\x08\x63ontract\x18\x0b \x03(\x0b\x32\x1e.protocol.Transaction.Contract\x12\x0f\n\x07scripts\x18\x0c \x01(\x0c\x12\x11\n\ttimestamp\x18\x0e \x01(\x03\x12\x11\n\tfee_limit\x18\x12 \x01(\x03\"\x80\x02\n\x0b\x42lockHeader\x12+\n\x08raw_data\x18\x01 \x01(\x0b\x32\x19.protocol.BlockHeader.raw\x12\x19\n\x11witness_signature\x18\x02 \x01(\x0c\x1a\xa8\x01\n\x03raw\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x12\n\ntxTrieRoot\x18\x02 \x01(\x0c\x12\x12\n\nparentHash\x18\x03 \x01(\x0c\x12\x0e\n\x06number\x18\x07 \x01(\x03\x12\x12\n\nwitness_id\x18\x08 \x01(\x03\x12\x17\n\x0fwitness_address\x18\t \x01(\x0c\x12\x0f\n\x07version\x18\n \x01(\x05\x12\x18\n\x10\x61\x63\x63ountStateRoot\x18\x0b \x01(\x0c\"a\n\x05\x42lock\x12+\n\x0ctransactions\x18\x01 \x03(\x0b\x32\x15.protocol.Transaction\x12+\n\x0c\x62lock_header\x18\x02 \x01(\x0b\x32\x15.protocol.BlockHeaderb\x06proto3')



_TRANSACTION = DESCRIPTOR.message_types_by_name['Transaction']
_TRANSACTION_CONTRACT = _TRANSACTION.nested_types_by_name['Contract']
_TRANSACTION_RESULT = _TRANSACTION.nested_types_by_name['Result']
_TRANSACTION_RAW = _TRANSACTION.nested_types_by_name['raw']
_BLOCKHEADER = DESCRIPTOR.message_types_by_name['BlockHeader']
_BLOCKHEADER_RAW = _BLOCKHEADER.nested_types_by_name['raw']
_BLOCK = DESCRIPTOR.message_types_by_name['Block']
_TRANSACTION_CONTRACT_CONTRACTTYPE = _TRANSACTION_CONTRACT.enum_types_by_name['ContractType']
_TRANSACTION_RESULT_CODE = _TRANSACTION_RESULT.enum_types_by_name['code']
_TRANSACTION_RESULT_CONTRACTRESULT = _TRANSACTION_RESULT.enum_types_by_name['contractResult']
Transaction = _reflection.GeneratedProtocolMessageType('Transaction', (_message.Message,), {

  'Contract' : _reflection.GeneratedProtocolMessageType('Contract', (_message.Message,), {
    'DESCRIPTOR' : _TRANSACTION_CONTRACT,
    '__module__' : 'graphsenselib.ingest.tron.grpc.core.chain_pb2'
    # @@protoc_insertion_point(class_scope:protocol.Transaction.Contract)
    })
  ,

  'Result' : _reflection.GeneratedProtocolMessageType('Result', (_message.Message,), {
    'DESCRIPTOR' : _TRANSACTION_RESULT,
    '__module__' : 'graphsenselib.ingest.tron.grpc.core.chain_pb2'
    # @@protoc_insertion_point(class_scope:protocol.Transaction.Result)
    })
  ,

  'raw' : _reflection.GeneratedProtocolMessageType('raw', (_message.Message,), {
    'DESCRIPTOR' : _TRANSACTION_RAW,
    '__module__' : 'graphsenselib.ingest.tron.grpc.core.chain_pb2'
    # @@protoc_insertion_point(class_scope:protocol.Transaction.raw)
    })
  ,
  'DESCRIPTOR' : _TRANSACTION,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.chain_pb2'
  # @@protoc_insertion_point(class_scope:protocol.Transaction)
  })
_sym_db.RegisterMessage(Transaction)
_sym_db.RegisterMessage(Transaction.Contract)
_sym_db.RegisterMessage(Transaction.Result)
_sym_db.RegisterMessage(Transaction.raw)

BlockHeader = _reflection.GeneratedProtocolMessageType('BlockHeader', (_message.Message,), {

  'raw' : _reflection.GeneratedProtocolMessageType('raw', (_message.Message,), {
    'DESCRIPTOR' : _BLOCKHEADER_RAW,
    '__module__' : 'graphsenselib.ingest.tron.grpc.core.chain_pb2'
    # @@protoc_insertion_point(class_scope:protocol.BlockHeader.raw)
    })
  ,
  'DESCRIPTOR' : _BLOCKHEADER,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.chain_pb2'
  # @@protoc_insertion_point(class_scope:protocol.BlockHeader)
  })
_sym_db.RegisterMessage(BlockHeader)
_sym_db.RegisterMessage(BlockHeader.raw)

Block = _reflection.GeneratedProtocolMessageType('Block', (_message.Message,), {
  'DESCRIPTOR' : _BLOCK,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.chain_pb2'
  # @@protoc_insertion_point(class_scope:protocol.Block)
  })
_sym_db.RegisterMessage(Block)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TRANSACTION._serialized_start=139
  _TRANSACTION._serialized_end=2383
  _TRANSACTION_CONTRACT._serialized_start=262
  _TRANSACTION_CONTRACT._serialized_end=1407
  _TRANSACTION_CONTRACT_CONTRACTTYPE._serialized_start=438
  _TRANSACTION_CONTRACT_CONTRACTTYPE._serialized_end=1407
  _TRANSACTION_RESULT._serialized_start=1410
  _TRANSACTION_RESULT._serialized_end=2128
  _TRANSACTION_RESULT_CODE._serialized_start=1790
  _TRANSACTION_RESULT_CODE._serialized_end=1820
  _TRANSACTION_RESULT_CONTRACTRESULT._serialized_start=1823
  _TRANSACTION_RESULT_CONTRACTRESULT._serialized_end=2128
  _TRANSACTION_RAW._serialized_start=2131
  _TRANSACTION_RAW._serialized_end=2383
  _BLOCKHEADER._serialized_start=2386
  _BLOCKHEADER._serialized_end=2642
  _BLOCKHEADER_RAW._serialized_start=2474
  _BLOCKHEADER_RAW._serialized_end=2642
  _BLOCK._serialized_start=2644
  _BLOCK._serialized_end=2741
# @@protoc_insertion_point(module_scope)



import click

from .accountsetl import convert_etl_to_ingestable_logs


@click.group()
def convert_cli():
    pass


@convert_cli.group()
def convert():
    """Useful file convertions tools for the graphsense infrastructure."""
    pass


@convert.command("log-to-ingestable-log")
@click.argument("inputfile", type=click.Path(exists=True))
@click.option("--prefix", type=str, default="ingestable")
def state(inputfile, prefix):
    """Summary
    Converts a log (events) file as exported by the ethereum etl component to
    a format that can be easily ingested in the log table using dsbulk.
    See schema/resources for the exact schema of the log table.
    \f
    Args:
        file (str): file to convert
        prefix (str): prefix for teh newly created file.
    """
    convert_etl_to_ingestable_logs(inputfile, prefix)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graphsenselib/ingest/tron/grpc/core/common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0graphsenselib/ingest/tron/grpc/core/common.proto\x12\x08protocol\"*\n\tAccountId\x12\x0c\n\x04name\x18\x01 \x01(\x0c\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0c\"J\n\tauthority\x12$\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x13.protocol.AccountId\x12\x17\n\x0fpermission_name\x18\x02 \x01(\x0c\"&\n\x03Key\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x0e\n\x06weight\x18\x02 \x01(\x03\"\xf1\x01\n\nPermission\x12\x31\n\x04type\x18\x01 \x01(\x0e\x32#.protocol.Permission.PermissionType\x12\n\n\x02id\x18\x02 \x01(\x05\x12\x17\n\x0fpermission_name\x18\x03 \x01(\t\x12\x11\n\tthreshold\x18\x04 \x01(\x03\x12\x11\n\tparent_id\x18\x05 \x01(\x05\x12\x12\n\noperations\x18\x06 \x01(\x0c\x12\x1b\n\x04keys\x18\x07 \x03(\x0b\x32\r.protocol.Key\"4\n\x0ePermissionType\x12\t\n\x05Owner\x10\x00\x12\x0b\n\x07Witness\x10\x01\x12\n\n\x06\x41\x63tive\x10\x02\"\x83\x07\n\rSmartContract\x12\x16\n\x0eorigin_address\x18\x01 \x01(\x0c\x12\x18\n\x10\x63ontract_address\x18\x02 \x01(\x0c\x12(\n\x03\x61\x62i\x18\x03 \x01(\x0b\x32\x1b.protocol.SmartContract.ABI\x12\x10\n\x08\x62ytecode\x18\x04 \x01(\x0c\x12\x12\n\ncall_value\x18\x05 \x01(\x03\x12%\n\x1d\x63onsume_user_resource_percent\x18\x06 \x01(\x03\x12\x0c\n\x04name\x18\x07 \x01(\t\x12\x1b\n\x13origin_energy_limit\x18\x08 \x01(\x03\x12\x11\n\tcode_hash\x18\t \x01(\x0c\x12\x10\n\x08trx_hash\x18\n \x01(\x0c\x1a\xf8\x04\n\x03\x41\x42I\x12\x31\n\x06\x65ntrys\x18\x01 \x03(\x0b\x32!.protocol.SmartContract.ABI.Entry\x1a\xbd\x04\n\x05\x45ntry\x12\x11\n\tanonymous\x18\x01 \x01(\x08\x12\x10\n\x08\x63onstant\x18\x02 \x01(\x08\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x37\n\x06inputs\x18\x04 \x03(\x0b\x32\'.protocol.SmartContract.ABI.Entry.Param\x12\x38\n\x07outputs\x18\x05 \x03(\x0b\x32\'.protocol.SmartContract.ABI.Entry.Param\x12\x39\n\x04type\x18\x06 \x01(\x0e\x32+.protocol.SmartContract.ABI.Entry.EntryType\x12\x0f\n\x07payable\x18\x07 \x01(\x08\x12N\n\x0fstateMutability\x18\x08 \x01(\x0e\x32\x35.protocol.SmartContract.ABI.Entry.StateMutabilityType\x1a\x34\n\x05Param\x12\x0f\n\x07indexed\x18\x01 \x01(\x08\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\"Y\n\tEntryType\x12\x14\n\x10UnknownEntryType\x10\x00\x12\x0f\n\x0b\x43onstructor\x10\x01\x12\x0c\n\x08\x46unction\x10\x02\x12\t\n\x05\x45vent\x10\x03\x12\x0c\n\x08\x46\x61llback\x10\x04\"a\n\x13StateMutabilityType\x12\x19\n\x15UnknownMutabilityType\x10\x00\x12\x08\n\x04Pure\x10\x01\x12\x08\n\x04View\x10\x02\x12\x0e\n\nNonpayable\x10\x03\x12\x0b\n\x07Payable\x10\x04\"0\n\x04Vote\x12\x14\n\x0cvote_address\x18\x01 \x01(\x0c\x12\x12\n\nvote_count\x18\x02 \x01(\x03\"I\n\x04Note\x12\r\n\x05value\x18\x01 \x01(\x03\x12\x17\n\x0fpayment_address\x18\x02 \x01(\t\x12\x0b\n\x03rcm\x18\x03 \x01(\x0c\x12\x0c\n\x04memo\x18\x04 \x01(\x0c*)\n\x0cResourceCode\x12\r\n\tBANDWIDTH\x10\x00\x12\n\n\x06\x45NERGY\x10\x01*7\n\x0b\x41\x63\x63ountType\x12\n\n\x06Normal\x10\x00\x12\x0e\n\nAssetIssue\x10\x01\x12\x0c\n\x08\x43ontract\x10\x02\x62\x06proto3')

_RESOURCECODE = DESCRIPTOR.enum_types_by_name['ResourceCode']
ResourceCode = enum_type_wrapper.EnumTypeWrapper(_RESOURCECODE)
_ACCOUNTTYPE = DESCRIPTOR.enum_types_by_name['AccountType']
AccountType = enum_type_wrapper.EnumTypeWrapper(_ACCOUNTTYPE)
BANDWIDTH = 0
ENERGY = 1
Normal = 0
AssetIssue = 1
Contract = 2


_ACCOUNTID = DESCRIPTOR.message_types_by_name['AccountId']
_AUTHORITY = DESCRIPTOR.message_types_by_name['authority']
_KEY = DESCRIPTOR.message_types_by_name['Key']
_PERMISSION = DESCRIPTOR.message_types_by_name['Permission']
_SMARTCONTRACT = DESCRIPTOR.message_types_by_name['SmartContract']
_SMARTCONTRACT_ABI = _SMARTCONTRACT.nested_types_by_name['ABI']
_SMARTCONTRACT_ABI_ENTRY = _SMARTCONTRACT_ABI.nested_types_by_name['Entry']
_SMARTCONTRACT_ABI_ENTRY_PARAM = _SMARTCONTRACT_ABI_ENTRY.nested_types_by_name['Param']
_VOTE = DESCRIPTOR.message_types_by_name['Vote']
_NOTE = DESCRIPTOR.message_types_by_name['Note']
_PERMISSION_PERMISSIONTYPE = _PERMISSION.enum_types_by_name['PermissionType']
_SMARTCONTRACT_ABI_ENTRY_ENTRYTYPE = _SMARTCONTRACT_ABI_ENTRY.enum_types_by_name['EntryType']
_SMARTCONTRACT_ABI_ENTRY_STATEMUTABILITYTYPE = _SMARTCONTRACT_ABI_ENTRY.enum_types_by_name['StateMutabilityType']
AccountId = _reflection.GeneratedProtocolMessageType('AccountId', (_message.Message,), {
  'DESCRIPTOR' : _ACCOUNTID,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
  # @@protoc_insertion_point(class_scope:protocol.AccountId)
  })
_sym_db.RegisterMessage(AccountId)

authority = _reflection.GeneratedProtocolMessageType('authority', (_message.Message,), {
  'DESCRIPTOR' : _AUTHORITY,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
  # @@protoc_insertion_point(class_scope:protocol.authority)
  })
_sym_db.RegisterMessage(authority)

Key = _reflection.GeneratedProtocolMessageType('Key', (_message.Message,), {
  'DESCRIPTOR' : _KEY,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
  # @@protoc_insertion_point(class_scope:protocol.Key)
  })
_sym_db.RegisterMessage(Key)

Permission = _reflection.GeneratedProtocolMessageType('Permission', (_message.Message,), {
  'DESCRIPTOR' : _PERMISSION,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
  # @@protoc_insertion_point(class_scope:protocol.Permission)
  })
_sym_db.RegisterMessage(Permission)

SmartContract = _reflection.GeneratedProtocolMessageType('SmartContract', (_message.Message,), {

  'ABI' : _reflection.GeneratedProtocolMessageType('ABI', (_message.Message,), {

    'Entry' : _reflection.GeneratedProtocolMessageType('Entry', (_message.Message,), {

      'Param' : _reflection.GeneratedProtocolMessageType('Param', (_message.Message,), {
        'DESCRIPTOR' : _SMARTCONTRACT_ABI_ENTRY_PARAM,
        '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
        # @@protoc_insertion_point(class_scope:protocol.SmartContract.ABI.Entry.Param)
        })
      ,
      'DESCRIPTOR' : _SMARTCONTRACT_ABI_ENTRY,
      '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
      # @@protoc_insertion_point(class_scope:protocol.SmartContract.ABI.Entry)
      })
    ,
    'DESCRIPTOR' : _SMARTCONTRACT_ABI,
    '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
    # @@protoc_insertion_point(class_scope:protocol.SmartContract.ABI)
    })
  ,
  'DESCRIPTOR' : _SMARTCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
  # @@protoc_insertion_point(class_scope:protocol.SmartContract)
  })
_sym_db.RegisterMessage(SmartContract)
_sym_db.RegisterMessage(SmartContract.ABI)
_sym_db.RegisterMessage(SmartContract.ABI.Entry)
_sym_db.RegisterMessage(SmartContract.ABI.Entry.Param)

Vote = _reflection.GeneratedProtocolMessageType('Vote', (_message.Message,), {
  'DESCRIPTOR' : _VOTE,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
  # @@protoc_insertion_point(class_scope:protocol.Vote)
  })
_sym_db.RegisterMessage(Vote)

Note = _reflection.GeneratedProtocolMessageType('Note', (_message.Message,), {
  'DESCRIPTOR' : _NOTE,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.common_pb2'
  # @@protoc_insertion_point(class_scope:protocol.Note)
  })
_sym_db.RegisterMessage(Note)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _RESOURCECODE._serialized_start=1493
  _RESOURCECODE._serialized_end=1534
  _ACCOUNTTYPE._serialized_start=1536
  _ACCOUNTTYPE._serialized_end=1591
  _ACCOUNTID._serialized_start=62
  _ACCOUNTID._serialized_end=104
  _AUTHORITY._serialized_start=106
  _AUTHORITY._serialized_end=180
  _KEY._serialized_start=182
  _KEY._serialized_end=220
  _PERMISSION._serialized_start=223
  _PERMISSION._serialized_end=464
  _PERMISSION_PERMISSIONTYPE._serialized_start=412
  _PERMISSION_PERMISSIONTYPE._serialized_end=464
  _SMARTCONTRACT._serialized_start=467
  _SMARTCONTRACT._serialized_end=1366
  _SMARTCONTRACT_ABI._serialized_start=734
  _SMARTCONTRACT_ABI._serialized_end=1366
  _SMARTCONTRACT_ABI_ENTRY._serialized_start=793
  _SMARTCONTRACT_ABI_ENTRY._serialized_end=1366
  _SMARTCONTRACT_ABI_ENTRY_PARAM._serialized_start=1124
  _SMARTCONTRACT_ABI_ENTRY_PARAM._serialized_end=1176
  _SMARTCONTRACT_ABI_ENTRY_ENTRYTYPE._serialized_start=1178
  _SMARTCONTRACT_ABI_ENTRY_ENTRYTYPE._serialized_end=1267
  _SMARTCONTRACT_ABI_ENTRY_STATEMUTABILITYTYPE._serialized_start=1269
  _SMARTCONTRACT_ABI_ENTRY_STATEMUTABILITYTYPE._serialized_end=1366
  _VOTE._serialized_start=1368
  _VOTE._serialized_end=1416
  _NOTE._serialized_start=1418
  _NOTE._serialized_end=1491
# @@protoc_insertion_point(module_scope)



import re
from datetime import datetime

from rich import print as printRich

from ..datatypes.abi import decode_db_logs


def log_sort_key(dlog, log):
    return (log.block_id, log.transaction_index, log.log_index)


def is_matching_row(query_terms, dlog_str, dlog_str_rep):
    return all(
        query.lower() in dlog_str.lower() or query.lower() in dlog_str_rep.lower()
        for query in query_terms
    )


def decoded_log_to_str(decoded_log) -> str:
    name = decoded_log["name"]
    addr = decoded_log["address"].lower()
    params = ",".join([f"{x['name']}={x['value']}" for x in decoded_log["data"]])
    return f"{addr}|{name}|{params}".replace("\n", "")


def make_replacer(**replacements):
    if len(replacements) != 0:
        locator = re.compile("|".join(re.escape(s) for s in replacements))

        def _doreplace(mo):
            return replacements[mo.group()]

        def replace(s):
            return locator.sub(_doreplace, s)

    else:

        def replace(s):
            return s

    return replace


def fetch_data(db, b, bts, contract, replace_names, query_terms, rowfn):
    if b not in bts:
        bts[b] = db.raw.get_block_timestamp(b)

    dtstring = bts[b].isoformat()
    decode_logs = decode_db_logs(db.raw.get_logs_in_block(b, contract=contract))

    decode_logs = sorted(
        [(dlog, log) for dlog, log in decode_logs],
        key=lambda x: log_sort_key(x[0], x[1]),
    )

    for dlog, log in decode_logs:
        # c = "0x" + log.address.hex()
        dlog_str_original = decoded_log_to_str(dlog)
        dlog_str = (
            f"{dtstring}|{b}|{log.log_index}|0x{log.tx_hash.hex()}|{dlog_str_original}"
        )
        dlog_str_rep = replace_names(dlog_str)
        if is_matching_row(query_terms, dlog_str, dlog_str_rep):
            rowfn(dlog, log, dlog_str_original, dlog_str, dlog_str_rep)


def trace(db, query_terms, contract, start, end, names, output_format):
    replace_names = make_replacer(**{k.lower(): v for k, v in names.items()})
    tx_mode = False
    if len(query_terms) == 1 and len(query_terms[0]) == 66:
        tx = query_terms[0]
        tx_data = db.raw.get_tx(tx)
        if tx_data is None:
            printRich("tx not found")
            return
        startblock = tx_data.block_id
        endblock = tx_data.block_id
        tx_mode = True

    if start is not None and end is not None:
        if start.isdigit():
            startblock = start
        else:
            startDt = datetime.fromisoformat(start)
            startblock = db.raw.find_block_nr_for_date(startDt)

        if end.isdigit():
            endblock = end
        else:
            endDt = datetime.fromisoformat(end)
            endblock = db.raw.find_block_nr_for_date(endDt)

    bts = {}

    if not tx_mode:
        from rich.progress import Progress

        with Progress() as progress:
            task1 = progress.add_task("[red]Searching...", total=endblock - startblock)
            for b in range(startblock, endblock + 1):
                progress.update(task1, advance=1)
                fetch_data(
                    db,
                    b,
                    bts,
                    contract,
                    replace_names,
                    query_terms,
                    lambda _, _1, _2, dlog_str, _3: printRich(dlog_str),
                )
    else:

        tx_mode_rows = []

        def append_row(dlog, log, dlog_str_original, dlog_str, dlog_str_rep):
            dlog_str = f"{log.log_index}|{replace_names(dlog_str_original)}"
            if output_format == "table":
                x = dlog_str.split("|")
                x[3] = x[3].replace(",", "\n")
                tx_mode_rows.append(x)
            else:
                printRich(dlog_str)

        for b in range(startblock, endblock + 1):
            fetch_data(db, b, bts, contract, replace_names, query_terms, append_row)

        if tx_mode and output_format == "table":
            from rich.console import Console
            from rich.table import Table

            dt = datetime.fromtimestamp(tx_data.block_timestamp)
            title = replace_names(
                f"{tx}: 0x{tx_data.from_address.hex()} "
                f"-> 0x{tx_data.to_address.hex()} @ {dt.isoformat()}"
            )

            table = Table(title=title, highlight=True)

            table.add_column("index", justify="right", style="cyan", no_wrap=True)
            table.add_column("project", style="magenta")
            table.add_column("event", style="red")
            table.add_column("params", justify="left", max_width=100)

            for x in tx_mode_rows:
                table.add_row(*x)

            console = Console()
            console.print(table)



import csv
from io import StringIO

import click

from ..cli.common import out_file, require_currency, require_environment
from ..config import config
from ..utils.console import console
from .monitoring import DbSummaryRecord, get_db_summary_record
from .notifications import send_msg_to_topic


@click.group()
def monitoring_cli():
    pass


@monitoring_cli.group("monitoring")
def monitoring():
    """Tools to monitor the graphsense infrastructure."""
    pass


@monitoring.command(
    "get-summary",
    short_help="Receives a summary record of the current database state.",
)
@require_environment()
@require_currency(required=False)
@out_file(required=False, append=True)
@click.option("--no-header/--header", default=False)
def summary(env, currency, out_file, no_header):
    """Receives a summary record of the current database state.
    \f

    Args:
        env (str): Env to work on
        currency (str): currency to work on (optional)
    """

    if currency is None:
        # create rows for all configured currencies
        records = [
            get_db_summary_record(env, currency_config)
            for currency_config in config.get_environment(
                env
            ).get_configured_currencies()
        ]
    else:
        records = [get_db_summary_record(env, currency)]

    output_stream = StringIO() if out_file is None else out_file

    writer = csv.DictWriter(output_stream, fieldnames=DbSummaryRecord.get_fields())

    if not no_header and (out_file is None or out_file.tell() == 0):
        # write header either for new file or if print on console
        writer.writeheader()

    writer.writerows([record.get_dict() for record in records])

    if out_file is None:
        # print records to stdout if no file is specified
        console.print(output_stream.getvalue())


@monitoring.command(
    "notify",
    short_help="Sends a message to the "
    "configured handlers (e.g. a slack channel) by topic.",
)
@click.option(
    "--topic",
    "-t",
    type=str,
    help="Topic to send to.",
    required=True,
)
@click.option(
    "--msg",
    "-m",
    type=str,
    help="Message to send.",
    required=True,
)
def notify(topic, msg):
    send_msg_to_topic(topic, msg)



import json
import logging
from datetime import date, datetime, timedelta
from typing import List, Optional

import pandas as pd
import requests

from ..db import DbFactory
from ..db.analytics import DATE_FORMAT

logger = logging.getLogger(__name__)

MIN_START = "2015-01-01"
FX_URL = "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-hist.zip"


def fetch_ecb_rates(symbol_list: List) -> pd.DataFrame:
    """Fetch and preprocess FX rates from ECB."""

    logger.info(f"Fetching conversion rates for FIAT currencies from {FX_URL}")
    rates_eur = pd.read_csv(FX_URL)  # exchange rates based on EUR
    rates_eur = rates_eur.iloc[:, :-1]  # remove empty last column
    rates_eur["EUR"] = 1.0
    # convert to values based on USD
    rates_usd = rates_eur[symbol_list].div(rates_eur.USD, axis=0)
    rates_usd["date"] = rates_eur.Date
    logger.info(f"Last record: {rates_usd.date.tolist()[0]}")
    return rates_usd


def cmc_historical_url(symbol: str, start: date, end: date) -> str:
    """Returns URL for CoinMarketCap API request."""
    return (
        "https://pro-api.coinmarketcap.com/v1/cryptocurrency/ohlcv/"
        + f"historical?symbol={symbol}&convert=USD"
        + f"&time_start={start}&time_end={end}"
    )


def parse_cmc_historical_response(
    response: requests.Response,
) -> pd.DataFrame:
    """Parse historical exchange rates (JSON) from CoinMarketCap."""

    json_data = json.loads(response.content)

    if "data" in json_data and "quotes" in json_data["data"]:
        json_data = [
            [elem["time_close"][:10], elem["quote"]["USD"]["close"]]
            for elem in json_data["data"]["quotes"]
        ]
    else:
        logger.error("Error: Coinmarketcap did not return any quotes.")
        raise SystemExit(100)

    return pd.DataFrame(json_data, columns=["date", "USD"])


def fetch_cmc_rates(
    start: str, end: str, crypto_currency: str, api_key: str
) -> pd.DataFrame:
    """Fetch cryptocurrency exchange rates from CoinMarketCap."""

    user_agent = (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_1) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/87.0.4280.88 Safari/537.36"
    )
    headers = {"User-Agent": user_agent, "X-CMC_PRO_API_KEY": api_key}

    start_date = date.fromisoformat(start) + timedelta(days=-1)
    end_date = date.fromisoformat(end)
    url = cmc_historical_url(crypto_currency, start_date, end_date)

    logger.info(f"Fetching {crypto_currency} exchange rates from {url}")
    rsession = requests.Session()
    rsession.mount("https://", requests.adapters.HTTPAdapter(max_retries=5))
    response = rsession.get(url, headers=headers)
    cmc_rates = parse_cmc_historical_response(response)

    if len(cmc_rates) > 0:
        last_record = cmc_rates.date.tolist()[-1]
    else:
        last_record = None
    logger.info(f"Last record: {last_record}")
    return cmc_rates


def fetch(env, currency, fiat_currencies, start_date, end_date, api_key):
    with DbFactory().from_config(env, currency) as db:
        return fetch_impl(
            db,
            env,
            currency,
            fiat_currencies,
            start_date,
            end_date,
            None,
            start_date != MIN_START,
            False,
            True,
            api_key,
        )


def fetch_impl(
    db: Optional[object],
    env: Optional[str],
    currency: str,
    fiat_currencies: List[str],
    start_date: Optional[str],
    end_date: Optional[str],
    table: Optional[str],
    force: bool,
    dry_run: bool,
    abort_on_gaps: bool,
    api_key: str,
):
    if datetime.fromisoformat(start_date) < datetime.fromisoformat(MIN_START):
        start_date = MIN_START

    # query most recent data
    if not force and db:
        logger.info(f"Get last imported rate from {db.raw.get_keyspace()}")
        most_recent_date = db.raw.get_last_exchange_rate_date(table=table)
        if most_recent_date is not None:
            start_date = most_recent_date.strftime(DATE_FORMAT)

    logger.info(f"*** Fetch exchange rates for {currency} ***")
    logger.info(f"Start date: {start_date}")
    logger.info(f"End date: {end_date}")
    logger.info(f"Target fiat currencies: {fiat_currencies}")

    if datetime.fromisoformat(start_date) > datetime.fromisoformat(end_date):
        logger.error("Error: start date after end date.")
        raise SystemExit

    # fetch cryptocurrency exchange rates in USD
    cmc_rates = fetch_cmc_rates(start_date, end_date, currency, api_key)

    ecb_rates = fetch_ecb_rates(fiat_currencies)

    # query conversion rates and merge converted values in exchange rates
    exchange_rates = cmc_rates
    date_range = pd.date_range(
        date.fromisoformat(start_date), date.fromisoformat(end_date)
    )
    date_range = pd.DataFrame(date_range, columns=["date"])
    date_range = date_range["date"].dt.strftime("%Y-%m-%d")

    for fiat_currency in set(fiat_currencies) - {"USD"}:
        ecb_rate = ecb_rates[["date", fiat_currency]].rename(
            columns={fiat_currency: "fx_rate"}
        )
        merged_df = cmc_rates.merge(ecb_rate, on="date", how="left").merge(
            date_range, how="right"
        )

        # fill gaps over weekends
        merged_df["fx_rate"].fillna(method="ffill", inplace=True)
        merged_df["fx_rate"].fillna(method="bfill", inplace=True)

        if abort_on_gaps and merged_df["fx_rate"].isnull().values.any():
            logger.error(
                "Error: found missing values for currency "
                f"{fiat_currency}, aborting import. Probably a weekend."
            )
            logger.error(merged_df[merged_df["fx_rate"].isnull()])
            if not dry_run:
                # in case of dry run let it run
                # to see what would have been written to the db
                if len(merged_df[merged_df["fx_rate"].isnull()]) > 4:
                    # if missing more than 4 days, critical error
                    raise SystemExit(2)
                else:
                    raise SystemExit(15)
        merged_df[fiat_currency] = merged_df["USD"] * merged_df["fx_rate"]
        merged_df = merged_df[["date", fiat_currency]]
        exchange_rates = exchange_rates.merge(merged_df, on="date")

    return exchange_rates


def ingest(
    env,
    currency,
    fiat_currencies,
    start_date,
    end_date,
    table,
    force,
    dry_run,
    abort_on_gaps,
    api_key,
):
    if dry_run:
        logger.warning("This is a Dry-Run. Nothing will be written to the database!")
    with DbFactory().from_config(env, currency) as db:
        exchange_rates = fetch_impl(
            db,
            env,
            currency,
            fiat_currencies,
            start_date,
            end_date,
            table,
            force,
            dry_run,
            abort_on_gaps,
            api_key,
        )

        # insert final exchange rates into Cassandra
        if "USD" not in fiat_currencies:
            exchange_rates.drop("USD", axis=1, inplace=True)
        exchange_rates["fiat_values"] = exchange_rates.drop("date", axis=1).to_dict(
            orient="records"
        )
        exchange_rates.drop(fiat_currencies, axis=1, inplace=True)

        # insert exchange rates into Cassandra table
        if not dry_run:
            logger.info(f"Writing to keyspace {db.raw.get_keyspace()}")
            if len(exchange_rates) > 0:
                db.raw.ingest(table, exchange_rates.to_dict("records"))
                logger.info(f"Inserted rates for {len(exchange_rates)} days: ")
                logger.info(
                    f"{exchange_rates.iloc[0].date} - {exchange_rates.iloc[-1].date}"
                )
            else:
                logger.info("Nothing to insert.")
        else:
            logger.info(
                "Dry run: No data inserted. "
                f"Would have inserted {len(exchange_rates)} days."
            )
            logger.info(exchange_rates)



import json
import logging
import os
import sys
from contextlib import contextmanager
from typing import Dict, List

import click
import requests

from .errorhandling import get_exception_digest

logger = logging.getLogger(__name__)


class ClickSlackErrorNotificationContext:
    def __init__(self, webhooks: List[str]):
        self.hooks = webhooks

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        e = sys.exc_info()[1]
        if isinstance(e, click.exceptions.Exit) and (
            e.exit_code == 0 or e.exit_code == 911
        ):
            # this is how click communicated all is well
            return
        if isinstance(e, click.exceptions.ClickException):
            return
        if isinstance(e, SystemExit) and e.code is not None and (e.code > 10):
            # Exit code 911 is used when lockfile is already used in
            # delta updates we don't what notifications for that
            # exit codes are chosen such that only below 10 are critical
            return
        if isinstance(e, KeyboardInterrupt):
            return
        if e is not None:
            for hook in self.hooks:
                try:
                    ret = send_exception_digest_to_slack(e, hook)
                    if ret.status_code != 200:
                        logger.error(
                            f"Failed to send exception detail to slack - {e}. "
                            f"Got status code {ret}, url wrong?"
                        )
                except Exception as e:
                    logger.error(f"Failed to send exception detail to slack - {e}")
        return False


@contextmanager
def on_exception_notify_slack(webhooks: List[str]):
    try:
        yield
    except Exception as e:
        for hook in webhooks:
            try:
                ret = send_exception_digest_to_slack(e, hook)
                if ret.status_code != 200:
                    logger.error(
                        f"Failed to send exception detail to slack - {e}. "
                        f"Got status code {ret}, url wrong?"
                    )
            except Exception as e:
                logger.error(f"Failed to send exception detail to slack - {e}")
        raise e


def send_exception_digest_to_slack(ex, webhook: str):
    machine = os.uname()
    machine_str = f"{machine.nodename} ({machine.sysname})"
    return send_message_to_slack(
        f"{get_exception_digest(ex)} \n"
        f"in {' '.join(sys.argv)}.\n"
        f"on {machine_str}\n"
        "Check the logs for more detail.",
        webhook,
    )


def send_message_to_slack(msg: str, webhook: str):
    return send_payload_to_slack({"text": msg}, webhook)


def send_payload_to_slack(payload: Dict, webhook: str):
    """Send a payload to slack via a webhook

    Args:
        payload (dict): is teh
        webhook (str): Webhook url.

    Returns:
        HTTP response code, i.e. <Response [503]>
    """
    headers = {"Content-type": "application/json"}
    return requests.post(webhook, json.dumps(payload), headers=headers)



import csv
import gzip
import logging
import os.path
import sys

logger = logging.getLogger(__name__)

ingestable_logs_fields = [
    "block_id_group",
    "block_id",
    "block_hash",
    "address",
    "data",
    "topics",
    "topic0",
    "tx_hash",
    "log_index",
    "transaction_index",
]


def convert_etl_to_ingestable_logs(filename: str, outfile_suffix: str):
    """Converts a log file as exported by the ingest component into a
    dsbulk imporable format.

    Ingest can be invoked as follows:
        dsbulk load -c csv -h ip -k eth_raw -t log -url PATH_TO_FILES \
         --connector.csv.fileNamePattern '/logs_abc.csv.gz' \
         --connector.csv.compression gzip --connector.csv.recursive true \
         --log.verbosity 2 --connector.csv.delimiter "|" \
         --connector.csv.maxCharsPerColumn 100000 \
         --dsbulk.connector.csv.nullValue "" \
         --dsbulk.connector.csv.ignoreLeadingWhitespaces true \
         --dsbulk.connector.csv.ignoreTrailingWhitespaces true


    Args:
        filename (str): Description
        outfile_suffix (str): Description

    Raises:
        e: Description
    """

    # set max fieldsize to 1MB, some logs have big data fields.
    # Default 16kb fail on certain logs.
    csv.field_size_limit(8388608)
    outfile = f"{filename}.{outfile_suffix}"

    if os.path.exists(outfile):
        logger.warning(f"Output file {outfile} already exists. I am done here.")
        sys.exit(199)

    if not filename.endswith("csv.gz"):
        logger.warning(f"Output file {outfile} name does not end with csv.gz.")
        sys.exit(201)
    try:
        with gzip.open(filename, "rt") as f:
            r = csv.DictReader(f, delimiter=",", quotechar='"')
            with gzip.open(outfile, "wt") as wf:
                w = csv.DictWriter(
                    wf,
                    fieldnames=ingestable_logs_fields,
                    delimiter="|",
                    quoting=csv.QUOTE_NONE,
                    quotechar="",
                )
                w.writeheader()
                for row in r:
                    if "transaction_hash" in row:
                        row.pop("transaction_hash")

                    tpcs_str = row["topics"].strip()

                    tpcs = tpcs_str.split("|") if len(tpcs_str) > 0 else []

                    if "topic0" not in row:
                        row["topic0"] = tpcs[0] if len(tpcs) > 0 else None

                    qt = ",".join([f'"{t}"' for t in tpcs])
                    row["topics"] = f"[{qt}]"

                    w.writerow(row)

    except Exception as e:
        if os.path.exists(outfile):
            os.remove(outfile)
        logger.error(f"Caught exception removed outfile {outfile}.")
        raise e



import sys

import click

from ..config import (
    config,
    default_environments,
    schema_types,
    supported_base_currencies,
)
from ..utils.console import console


def require_environment(required=True):
    def inner(function):
        function = click.option(
            "--env",
            "-e",
            type=click.Choice(default_environments, case_sensitive=False),
            help="Environment to work on",
            required=required,
        )(function)
        return function

    return inner


def require_schema_type(required=True):
    def inner(function):
        function = click.option(
            "--schema",
            "-s",
            type=click.Choice(schema_types, case_sensitive=False),
            help="Type of schema supported by graphsense.",
            required=required,
        )(function)
        return function

    return inner


def require_currency(required=True):
    def inner(function):
        function = click.option(
            "--currency",
            "-c",
            type=click.Choice(supported_base_currencies, case_sensitive=False),
            help="Currency to work on",
            required=required,
        )(function)
        return function

    return inner


def out_file(required=True, append=False):
    def inner(function):
        function = click.option(
            "--out-file",
            "-o",
            type=click.File(
                mode="a" if append else "w",
                encoding=None,
                errors="strict",
                lazy=None,
                atomic=False,
            ),
            help="Output text file.",
            required=required,
        )(function)
        return function

    return inner


def try_load_config(filename: str):
    try:
        config.load(filename=filename)
        return config
    except Exception as e:
        console.print("There are errors in you graphsenselib config:")
        console.rule("Errors")
        console.print(e)
        console.rule("Suggestions")
        file_loc = " or ".join(config.Config.default_files)
        console.print(
            "Maybe there is no config file specified. "
            f"Please create one in {file_loc} or specify a custom config path"
            f" in the environment variable {config.Config.file_env_var}."
        )
        console.rule("Template")
        console.print(config.generate_yaml(DEBUG=False))
        sys.exit(10)



import logging
from abc import ABC, abstractmethod
from collections import OrderedDict
from collections.abc import Iterable
from datetime import datetime, timezone
from operator import itemgetter
from typing import Dict, List, Optional, Tuple

from bitcoinetl.enumeration.chain import Chain
from bitcoinetl.jobs.export_blocks_job import ExportBlocksJob
from bitcoinetl.rpc.bitcoin_rpc import BitcoinRpc
from bitcoinetl.service.btc_service import BtcService
from blockchainetl.jobs.exporters.in_memory_item_exporter import InMemoryItemExporter
from blockchainetl.thread_local_proxy import ThreadLocalProxy
from btcpy.structs.address import P2pkhAddress
from btcpy.structs.script import ScriptBuilder
from methodtools import lru_cache as mlru_cache

from ..config import GRAPHSENSE_DEFAULT_DATETIME_FORMAT, get_reorg_backoff_blocks
from ..db import AnalyticsDb
from ..utils import bytes_to_hex, flatten, hex_to_bytes, parse_timestamp, strip_0x
from ..utils.account import get_id_group
from ..utils.bch import bch_address_to_legacy
from ..utils.logging import suppress_log_level
from ..utils.signals import graceful_ctlc_shutdown
from .common import cassandra_ingest, write_to_sinks

TX_HASH_PREFIX_LENGTH = 5
TX_BUCKET_SIZE = 25_000
BLOCK_BUCKET_SIZE = 100

OUTPUTS_CACHE_ITEMS = 2**24  # approx 16 mio.


logger = logging.getLogger(__name__)

CHAIN_MAPPING = {
    "btc": Chain.BITCOIN,
    "ltc": Chain.LITECOIN,
    # Use the new api for btc (in btc etl jargon), this is
    # required for bitcoin cash, otherwise the index field in the transactions
    # is not filled correctly.
    "bch": Chain.BITCOIN,
    "zec": Chain.ZCASH,
}


class UnknownScriptType(Exception):
    pass


class UnknownAddressType(Exception):
    pass


class P2pkParserException(Exception):
    pass


class InputNotFoundException(Exception):
    pass


class BtcStreamerAdapter:
    def __init__(self, bitcoin_rpc, chain=Chain.BITCOIN, batch_size=2, max_workers=5):
        """Summary

        Args:
            bitcoin_rpc (TYPE): Description
            chain (TYPE, optional): Description
            batch_size (int, optional): Description
            max_workers (int, optional): Description
        """
        self.bitcoin_rpc = bitcoin_rpc
        self.chain = chain
        self.btc_service = BtcService(bitcoin_rpc, chain)
        self.batch_size = batch_size
        self.max_workers = max_workers

    def open(self):  # noqa
        self.item_exporter.open()

    def get_current_block_number(self):
        return int(self.btc_service.get_latest_block().number)

    def export_transactions(self, start_block, end_block):
        transactions_item_exporter = InMemoryItemExporter(item_types=["transaction"])

        transactions_job = ExportBlocksJob(
            start_block=start_block,
            end_block=end_block,
            batch_size=self.batch_size,
            bitcoin_rpc=self.bitcoin_rpc,
            max_workers=self.max_workers,
            item_exporter=transactions_item_exporter,
            chain=self.chain,
            export_blocks=False,
            export_transactions=True,
        )
        transactions_job.run()

        transactions = transactions_item_exporter.get_items("transaction")

        return transactions

    def export_blocks_and_transactions(self, start_block, end_block):
        blocks_and_transactions_item_exporter = InMemoryItemExporter(
            item_types=["block", "transaction"]
        )

        blocks_and_transactions_job = ExportBlocksJob(
            start_block=start_block,
            end_block=end_block,
            batch_size=self.batch_size,
            bitcoin_rpc=self.bitcoin_rpc,
            max_workers=self.max_workers,
            item_exporter=blocks_and_transactions_item_exporter,
            chain=self.chain,
            export_blocks=True,
            export_transactions=True,
        )
        blocks_and_transactions_job.run()

        blocks = blocks_and_transactions_item_exporter.get_items("block")
        transactions = blocks_and_transactions_item_exporter.get_items("transaction")

        return blocks, transactions

    def close(self):
        self.item_exporter.close()


class OutputResolverBase(ABC):
    @abstractmethod
    def get_output(self, tx_hash) -> Dict:
        pass

    @abstractmethod
    def add_output(self, tx_hash, output) -> None:
        pass

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class LRUCache(OrderedDict):
    def __init__(self, *args, **kwds):
        self.size_limit = kwds.pop("size_limit", None)
        OrderedDict.__init__(self, *args, **kwds)
        self._check_size_limit()

    def __setitem__(self, key, value):
        OrderedDict.__setitem__(self, key, value)
        self._check_size_limit()

    def _check_size_limit(self):
        if self.size_limit is not None:
            while len(self) > self.size_limit:
                self.popitem(last=False)


class CassandraOutputResolver(OutputResolverBase):
    """Output resolver that uses the gs-cassandra database to resolve
    spent inputs.

    Attributes:
        cache (TYPE): Description
        db (TYPE): Database connection
        tx_bucket_size (TYPE): Description
        tx_prefix_length (TYPE): Description
    """

    def __init__(
        self,
        db: AnalyticsDb,
        tx_bucket_size: int = None,
        tx_prefix_length: int = None,
    ):
        self.db = db
        self.tx_bucket_size = tx_bucket_size
        self.tx_prefix_length = tx_prefix_length
        self.cache = LRUCache(size_limit=OUTPUTS_CACHE_ITEMS)

    def __enter__(self):
        self.db.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.db.__exit__(exc_type, exc_val, exc_tb)

    @mlru_cache(maxsize=OUTPUTS_CACHE_ITEMS)
    def get_output(self, tx_hash) -> Dict:
        if tx_hash in self.cache.keys():
            return self.cache.get(tx_hash)
        return self._get_from_db(tx_hash)

    def _get_from_db(self, tx_hash):
        outputs = self.db.raw.get_tx_outputs(
            tx_hash,
            tx_bucket_size=self.tx_bucket_size,
            tx_prefix_length=self.tx_prefix_length,
        )
        return outputs

    def add_output(self, tx_hash, output) -> None:
        r = {k: output[k] for k in ["type", "addresses", "value"]}
        self.cache.setdefault(tx_hash, {}).setdefault(output["index"], r)

    def get_cache_stats(self) -> str:
        """Gets a string summary of the state of the cache

        Returns:
            str: String summary of cache
        """
        return (
            f"Frontend and DB cache: {self.get_output.cache_info()}, "
            f"Run Local Cache: {len(self.cache)} items"
        )


def get_last_block_yesterday(
    btc_adapter: BtcStreamerAdapter, last_synced_block: int
) -> int:
    until_date = datetime.utcnow().replace(
        hour=0, minute=0, second=0, microsecond=0, tzinfo=timezone.utc
    )
    until_timestamp = until_date.timestamp()

    block_id = last_synced_block
    block = btc_adapter.btc_service.get_block(block_id)
    while block.timestamp >= until_timestamp:
        block_id -= 1
        block = btc_adapter.btc_service.get_block(block_id)

    block_number = block.number
    timestamp = parse_timestamp(block.timestamp)
    logger.info(
        f"Determining latest block before {until_date.isoformat()} "
        f"its {block_number:,} at {timestamp.isoformat()}",
    )
    return block_number


def ingest_block_transactions(
    items: Iterable,
    db: AnalyticsDb,
    sink_config: dict,
    block_bucket_size: int,
) -> None:
    mapping = {}
    for tx in items:
        mapping.setdefault(tx["block_id"], []).append(tx)

    items = []
    for block, block_txs in mapping.items():
        items.append(
            {
                "block_id_group": get_id_group(block, block_bucket_size),
                "block_id": block,
                "txs": [tx_stats(x) for x in block_txs],
            }
        )

    write_to_sinks(db, sink_config, "block_transactions", items)


def ingest_blocks(
    items: Iterable,
    db: AnalyticsDb,
    sink_config: dict,
) -> None:
    write_to_sinks(db, sink_config, "block", items)


def ingest_tx_refs(
    items: Iterable,
    db: AnalyticsDb,
    sink_config: dict,
) -> None:
    write_to_sinks(db, sink_config, "transaction_spent_in", items)
    write_to_sinks(db, sink_config, "transaction_spending", items)


def prepare_blocks_inplace(blocks: Iterable, block_bucket_size: int) -> None:
    """Preprocesses a block and for ingestation into Cassandra.

    Args:
        blocks (Iterable): iterable of block structures
        block_bucket_size (int): size of a block bucket for cassandra partitioning
    """
    blob_columns = ["hash"]
    pop_columns = [
        "type",
        "size",
        "stripped_size",
        "weight",
        "version",
        "merkle_root",
        "nonce",
        "bits",
        "coinbase_param",
    ]
    for block in blocks:
        for i in pop_columns:
            block.pop(i)

        for elem in blob_columns:
            block[elem] = hex_to_bytes(
                block[elem]
            )  # convert hex strings to byte arrays (blob in Cassandra)

        block["block_id_group"] = get_id_group(block["number"], block_bucket_size)

        block["block_id"] = block.pop("number")
        block["block_hash"] = block.pop("hash")
        block["no_transactions"] = block.pop("transaction_count")


_address_types = {  # based on BlockSci values (type 0 .. 10)
    "nonstandard": 1,
    "pubkey": 2,
    "p2pk": 2,
    "pubkeyhash": 3,
    "p2pkh": 3,
    "multisig_pubkey": 4,
    "scripthash": 5,
    "p2sh": 5,
    "multisig": 6,
    "null": 7,
    "nulldata": 7,
    "witness_v0_keyhash": 8,
    "p2wpkhv0": 8,
    "witness_v0_scripthash": 9,
    "p2wshv0": 9,
    "witness_unknown": 10,
    "witness_v1_taproot": 11,
    "shielded": 12,
}


def addresstype_to_int(addr_type: str) -> int:
    """Convert btcpy address types to integers

    Args:
        addr_type (str): Description

    Returns:
        str: address type as integer

    Raises:
        UnknownAddressType: Description
    """
    if isinstance(addr_type, int):  # address_type has already been resolved
        return addr_type

    if addr_type in _address_types:
        return _address_types.get(addr_type)

    raise UnknownAddressType(f"unknown address type {addr_type}")


def address_as_string(x):
    if x["type"] in ["null", "nulldata", "nonstandard", "witness_unknown", "shielded"]:
        return None
    return x["addresses"]


def tx_io_summary(x):
    """Creates a short summary of in and outputs

    Args:
        x (TYPE): Description

    Returns:
        TYPE: list of form address, value, type
    """
    return [address_as_string(x), x["value"], addresstype_to_int(x["type"])]


def tx_stats(tx):
    return (
        tx["tx_id"],
        len(tx["inputs"]),
        len(tx["outputs"]),
        tx["total_input"],
        tx["total_output"],
    )


def parse_script(s: str) -> Tuple[List[str], str]:
    """Parses the output addresses from a bitcoin-like locking script

    Args:
        s (str): script in binary hex format

    Returns:
        Tuple[List[str], str]: address list and script type

    Raises:
        P2pkParserException: if P2PK script can't be parsed
        UnknownScriptType: On unknown script type
    """
    script = ScriptBuilder.identify(s)

    if script.type == "p2pk":
        try:
            return [str(P2pkhAddress(script.pubkey.hash(), mainnet=True))], script.type
        except Exception as e:
            raise P2pkParserException(
                f"ScriptParseError: cannot parse pubkey from {s}"
                f" (of type {script.type}) --- {e}"
            )

    if script.type == "p2pkh":
        return [str(P2pkhAddress(script.pubkeyhash, mainnet=True))], script.type

    if script.type == "multisig":
        return [
            str(P2pkhAddress(k.hash(), mainnet=True))
            for k in script.pubkeys
            if hasattr(k, "hash")
        ], script.type

    if script.type in ["p2sh", "p2wpkhv0", "p2wshv0"]:
        return [str(script.address(mainnet=True))], script.type

    if script.type == "nulldata":
        return None, script.type

    raise UnknownScriptType(
        f"ScriptParseError: not handling script type {script.type} at the moment."
    )


def enrich_txs(
    txs: Iterable, resolver: OutputResolverBase, ignore_missing_outputs: bool
) -> None:
    """Resolves transaction input links to the spent transactions.

    Args:
        txs (Iterable): transactions as produced by btc etl
        resolver (OutputResolverBase): instance of a
                                        resolver that finds the spent outputs
        ignore_missing_outputs (bool): if True skips outputs that can not be resolved

    Raises:
        InputNotFoundException: If inputs can not be
                                resolved and ignore_missing_outputs is false
    """

    # add outputs to cache before processing the inputs
    # help to avoid ordering related errors while resolving inputs
    # one instance where this happens is bitcoin cash;
    # btc-etl does not respect the new CTOR ordering for bch transactions.
    # given the returned tx order, tx inputs do not always resolve correctly.
    # an example is block (801379, tx at index 2 spends tx at index 12)
    # we circumvent this issue by pre-populating the cache.

    for tx in txs:
        # process outputs
        for o in tx["outputs"]:
            if o["addresses"]:
                if o["addresses"][0] and o["addresses"][0].startswith("bitcoincash:"):
                    o["addresses"] = [bch_address_to_legacy(a) for a in o["addresses"]]

                if o["addresses"][0] and o["addresses"][0].startswith("nonstandard"):
                    try:
                        address_list, scripttype = parse_script(o["script_hex"])
                        o["addresses"] = (
                            address_list if address_list else o["addresses"]
                        )
                        o["type"] = scripttype
                    except (
                        UnknownScriptType,
                        UnknownAddressType,
                        P2pkParserException,
                    ) as exception:
                        logger.warning(
                            f"{exception}: cannot parse output script {o}"
                            f" from tx {tx.get('hash')}"
                        )

                resolver.add_output(tx["hash"], o)

    for tx in txs:
        if not tx["is_coinbase"]:
            # process inputs
            for i in tx["inputs"]:
                if i["spent_transaction_hash"]:
                    ref, ind = (
                        i["spent_transaction_hash"],
                        i["spent_output_index"],
                    )
                    try:
                        resolved_outputs = resolver.get_output(ref)
                        resolved = (
                            resolved_outputs.get(ind) if resolved_outputs else None
                        )

                        if resolved is None and ignore_missing_outputs:
                            logger.warning("Could not resolve spent_txs outputs")
                            continue
                        elif resolved is None:
                            raise InputNotFoundException(
                                f"Spent Tx ({ref}) outputs " " not found"
                            )

                        i["addresses"] = resolved["addresses"]
                        i["type"] = resolved["type"]
                        i["value"] = resolved["value"]
                    except (
                        UnknownScriptType,
                        UnknownAddressType,
                        P2pkParserException,
                    ) as exception:
                        logger.warning(
                            f"tx input cannot be resolved for {i['addresses']}"
                        )
                        logger.warning(exception)

        tx["input_value"] = sum(
            [i["value"] for i in tx["inputs"] if i["value"] is not None]
        )


def prepare_transactions_inplace(
    txs: Iterable,
    next_tx_id: int,
    tx_hash_prefix_len: int,
    tx_bucket_size: int,
) -> None:
    pop_columns = [
        "type",
        "size",
        "virtual_size",
        "version",
        "lock_time",
        "index",
        "fee",
    ]
    blob_columns = ["hash"]

    assert all(tx["index"] is not None for tx in txs)

    txs = sorted(
        txs, key=itemgetter("block_number", "index")
    )  # because bitcoin-etl does not guarantee a sort order

    for tx in txs:
        for i in pop_columns:
            tx.pop(i)

        tx["tx_prefix"] = tx["hash"][:tx_hash_prefix_len]

        for elem in blob_columns:
            tx[elem] = hex_to_bytes(
                tx[elem]
            )  # convert hex strings to byte arrays (blob in Cassandra)

        tx["block_id"] = tx.pop("block_number")
        tx["coinbase"] = tx.pop("is_coinbase")
        tx["coinjoin"] = is_coinjoin(tx)

        tx["inputs_raw"] = tx["inputs"]
        tx["outputs_raw"] = tx["outputs"]

        tx["inputs"] = [tx_io_summary(x) for x in tx["inputs"]]
        tx["outputs"] = [tx_io_summary(x) for x in tx["outputs"]]
        tx["timestamp"] = tx.pop("block_timestamp")
        tx["total_input"] = tx.pop("input_value")
        tx["total_output"] = tx.pop("output_value")
        tx["tx_hash"] = tx.pop("hash")

        tx["tx_id_group"] = get_id_group(next_tx_id, tx_bucket_size)
        tx["tx_id"] = next_tx_id
        next_tx_id += 1


def get_tx_refs(spending_tx_hash: str, raw_inputs: Iterable, tx_hash_prefix_len: int):
    tx_refs = []
    spending_tx_hash = hex_to_bytes(spending_tx_hash)
    for inp in raw_inputs:
        spending_input_index = inp["index"]
        spent_tx_hash = hex_to_bytes(inp["spent_transaction_hash"])
        spent_output_index = inp["spent_output_index"]
        if spending_tx_hash is not None and spent_tx_hash is not None:
            # in zcash refs can be None in case of shielded txs.
            tx_refs.append(
                {
                    "spending_tx_hash": spending_tx_hash,
                    "spending_input_index": spending_input_index,
                    "spent_tx_hash": spent_tx_hash,
                    "spent_output_index": spent_output_index,
                    "spending_tx_prefix": strip_0x(bytes_to_hex(spending_tx_hash))[
                        :tx_hash_prefix_len
                    ],
                    "spent_tx_prefix": strip_0x(bytes_to_hex(spent_tx_hash))[
                        :tx_hash_prefix_len
                    ],
                }
            )

    return tx_refs


def ingest_transactions(
    items: Iterable,
    db: AnalyticsDb,
    sink_config: dict,
) -> None:
    write_to_sinks(db, sink_config, "transaction", items)


def ingest_transaction_lookups(
    items: Iterable,
    db: AnalyticsDb,
    sink_config: dict,
) -> None:
    res = [{k: tx[k] for k in ("tx_prefix", "tx_hash", "tx_id")} for tx in items]
    write_to_sinks(db, sink_config, "transaction_by_tx_prefix", res)


def is_coinjoin(tx) -> bool:
    # https://github.com/citp/BlockSci/blob/14ccc9358443b2eb5730bb2902c4b11ab7928abf/src/heuristics/tx_identification.cpp#L48
    if tx["input_count"] < 2 or tx["output_count"] < 3:
        return False

    # Each participant contributes a spend and a change output
    participant_count = int((tx["output_count"] + 1) / 2)
    if participant_count > tx["input_count"]:
        return False

    input_addresses = {
        str(x["addresses"]) for x in tx["inputs"] if "addresses" in x.keys()
    }

    if participant_count > len(input_addresses):
        return False

    # The most common output value should appear 'participantCount' times;
    # if multiple values are tied for 'most common', the lowest value is used
    output_values = {}
    for o in tx["outputs"]:
        output_values[o["value"]] = output_values.get(o["value"], 0) + 1

    highest_count = max(output_values.values())
    val, frequency = min(
        [(key, c) for key, c in output_values.items() if c == highest_count]
    )

    if frequency != participant_count:
        return False

    # Exclude transactions sending dust outputs (unlikely to be CoinJoin)
    if val in (546, 2730):
        return False

    return True


def print_block_info(
    last_synced_block: int, last_ingested_block: Optional[int]
) -> None:
    """Display information about number of synced/ingested blocks.

    Args:
        last_synced_block (int): Description
        last_ingested_block (Optional[int]): Description
    """

    logger.warning(f"Last synced block: {last_synced_block:,}")
    if last_ingested_block is None:
        logger.warning("Last ingested block: None")
    else:
        logger.warning(f"Last ingested block: {last_ingested_block:,}")


def ingest_summary_statistics_cassandra(
    db: AnalyticsDb,
    timestamp: int,
    total_blocks: int,
    total_txs: int,
) -> None:
    """Summary

    Args:
        db (AnalyticsDb): Description
        timestamp (int): Description
        total_blocks (int): Description
        total_txs (int): Description
    """
    cassandra_ingest(
        db,
        "summary_statistics",
        [
            {
                "id": db.raw.keyspace_name(),
                "timestamp": timestamp,
                "no_blocks": total_blocks,
                "no_txs": total_txs,
            }
        ],
    )


def ingest_configuration_cassandra(
    db: AnalyticsDb,
    block_bucket_size: int,
    tx_hash_prefix_len: int,
    tx_bucket_size: int,
) -> None:
    """Store configuration details in Cassandra table.

    Args:
        db (AnalyticsDb): Description
        block_bucket_size (int): Description
        tx_hash_prefix_len (int): Description
        tx_bucket_size (int): Description
    """
    cassandra_ingest(
        db,
        "configuration",
        [
            {
                "id": db.raw.keyspace_name(),
                "block_bucket_size": int(block_bucket_size),
                "tx_prefix_length": tx_hash_prefix_len,
                "tx_bucket_size": tx_bucket_size,
            }
        ],
    )


def get_connection_from_url(provider_uri: str) -> ThreadLocalProxy:
    return ThreadLocalProxy(lambda: BitcoinRpc(provider_uri))


def get_stream_adapter(
    currency: str, provider_uri: str, batch_size: int
) -> BtcStreamerAdapter:
    proxy = get_connection_from_url(provider_uri)
    return BtcStreamerAdapter(
        proxy, chain=CHAIN_MAPPING.get(currency, None), batch_size=batch_size
    )


def ingest(
    db: AnalyticsDb,
    currency: str,
    provider_uri: str,
    sink_config: dict,
    user_start_block: Optional[int],
    user_end_block: Optional[int],
    batch_size: int,
    info: bool,
    previous_day: bool,
    provider_timeout: int,
    mode: str,
):
    if currency not in CHAIN_MAPPING:
        raise ValueError(
            f"{currency} not supported by ingest module,"
            f" supported {list(CHAIN_MAPPING.keys())}"
        )

    import_refs = mode == "utxo_only_tx_graph" or mode == "utxo_with_tx_graph"
    import_base_data = mode != "utxo_only_tx_graph"

    del mode

    logger.info(
        f"Importing base data: {import_base_data}, import tx refs {import_refs}"
    )

    btc_adapter = get_stream_adapter(currency, provider_uri, batch_size=batch_size)

    resolver = CassandraOutputResolver(
        db,
        tx_bucket_size=TX_BUCKET_SIZE,
        tx_prefix_length=TX_HASH_PREFIX_LENGTH,
    )

    last_synced_block = btc_adapter.get_current_block_number()
    last_ingested_block = db.raw.get_highest_block()
    print_block_info(last_synced_block, last_ingested_block)

    start_block = 0
    if user_start_block is None:
        if last_ingested_block is not None:
            start_block = last_ingested_block + 1
    else:
        start_block = user_start_block

    end_block = last_synced_block - get_reorg_backoff_blocks(currency)
    if user_end_block is not None:
        end_block = user_end_block

    if previous_day:
        end_block = get_last_block_yesterday(btc_adapter, last_synced_block)

    if start_block > end_block:
        print("No blocks to ingest")
        return

    # if info then only print block info and exit
    if info:
        logger.info(
            f"Would ingest block range "
            f"{start_block:,} - {end_block:,} ({end_block-start_block:,} blks) "
            f"into {list(sink_config.keys())} "
        )
        return

    time1 = datetime.now()
    count = 0

    logger.info(
        f"Ingesting block range "
        f"{start_block:,} - {end_block:,} ({end_block-start_block:,} blks) "
        f"into {list(sink_config.keys())} "
    )

    if import_base_data:
        ingest_configuration_cassandra(
            db,
            block_bucket_size=BLOCK_BUCKET_SIZE,
            tx_hash_prefix_len=TX_HASH_PREFIX_LENGTH,
            tx_bucket_size=TX_BUCKET_SIZE,
        )

    with graceful_ctlc_shutdown() as check_shutdown_initialized:
        for block_id in range(start_block, end_block + 1, batch_size):
            current_end_block = min(end_block, block_id + batch_size - 1)

            with suppress_log_level(logging.INFO):
                blocks, txs = btc_adapter.export_blocks_and_transactions(
                    block_id, current_end_block
                )

            tx_refs = flatten(
                [
                    get_tx_refs(tx["hash"], tx["inputs"], TX_HASH_PREFIX_LENGTH)
                    for tx in txs
                ]
            )

            prepare_blocks_inplace(blocks, BLOCK_BUCKET_SIZE)

            if import_base_data:
                # until bitcoin-etl progresses
                # with https://github.com/blockchain-etl/bitcoin-etl/issues/43
                enrich_txs(txs, resolver, ignore_missing_outputs=False)

                latest_tx_id = db.raw.get_latest_tx_id_before_block(block_id)

                prepare_transactions_inplace(
                    txs, latest_tx_id + 1, TX_HASH_PREFIX_LENGTH, TX_BUCKET_SIZE
                )

                ingest_blocks(blocks, db, sink_config)
                ingest_transaction_lookups(
                    txs,
                    db,
                    sink_config,
                )
                ingest_transactions(txs, db, sink_config)
                ingest_block_transactions(txs, db, sink_config, BLOCK_BUCKET_SIZE)

            if import_refs:
                ingest_tx_refs(tx_refs, db, sink_config)

            last_block = blocks[-1]
            last_block_ts = last_block["timestamp"]
            last_block_id = last_block["block_id"]
            count += batch_size

            if count % 100 == 0:
                last_block_date = parse_timestamp(last_block_ts)
                time2 = datetime.now()
                time_delta = (time2 - time1).total_seconds()
                logger.info(
                    f"Last processed block: {current_end_block:,} "
                    f"[{last_block_date.strftime(GRAPHSENSE_DEFAULT_DATETIME_FORMAT)}] "
                    f"({count/time_delta:.1f} blocks/s)"
                )
                logger.debug(resolver.get_cache_stats())
                time1 = time2
                count = 0

            if check_shutdown_initialized():
                break

    last_block_date = parse_timestamp(last_block_ts)
    logger.info(
        f"Processed block range {start_block:,} - {last_block_id:,} "
        f" ({last_block_date.strftime(GRAPHSENSE_DEFAULT_DATETIME_FORMAT)})"
    )

    # store configuration details
    if "cassandra" in sink_config.keys() and import_base_data:
        last_block = blocks[-1]
        last_tx = txs[-1]
        ingest_summary_statistics_cassandra(
            db,
            timestamp=last_block_ts,
            total_blocks=last_block_id + 1,
            total_txs=last_tx["tx_id"] + 1,
        )



# -*- coding: utf-8 -*-
import pytest

from graphsenselib.ingest.utxo import (
    P2pkParserException,
    UnknownAddressType,
    UnknownScriptType,
    address_as_string,
    addresstype_to_int,
    parse_script,
)


def test_pubkey():
    scripthex = "4104a9d6840fdd1497b3067b8066db783acf90bf42071a38fe2cf6d2d8a04835d0b5c45716d8d6012ab5d56c7824c39718f7bc7486d389cd0047f53785f9a63c0c9dac"  # noqa
    adress_list, script_type = parse_script(scripthex)
    assert script_type == "p2pk"
    assert adress_list == ["1ND5TQ2AnQatxqmHrdMT8utb3aWgLKsc9w"]


def test_pubkeyhash():
    scripthex = "76a91455240639b8ef1882127252c12f90354cc2fa85c088ac"
    adress_list, script_type = parse_script(scripthex)
    assert script_type == "p2pkh"
    assert adress_list == ["18mBbKWTGAbdbP2h24Nbx5AtjWLxScJP94"]


def test_scripthash():
    scripthex = "a91475b18a0606d57fb14a607e04ff6bcd81db44639187"
    adress_list, script_type = parse_script(scripthex)
    assert script_type == "p2sh"
    assert adress_list == ["3CRKfdBV78DD2R44SrbySmWvC9VdRRdHxQ"]


def test_multisig():
    scripthex = "514104d68bea20391d02ce6d668b1bf811f40954d9a8a1eb0c7087ddf33b5f4ad95b3c00c7542006dd4d8654f4e2975e59cc3203057f3ed8dc1a859c4d6e64dc841f0921028b6e78f41eff35d18875b5c450c62d1963792c7288e0ed40f63f258f0f633eaa52ae"  # noqa
    adress_list, script_type = parse_script(scripthex)
    assert script_type == "multisig"
    assert adress_list == [
        "19zUcPa1ENfnSy8RMJ3AbF4TXMaZrskKKb",
        "1C3gCLqhxPb8asjtuLBe6dF5FzXRo5nANe",
    ]


def test_witness_v0_keyhash():
    scripthex = "0014f9da60876f467c88ccd5317f5fc21d124b2538c2"
    adress_list, script_type = parse_script(scripthex)
    assert script_type == "p2wpkhv0"
    assert adress_list == ["bc1ql8dxppm0ge7g3nx4x9l4lssazf9j2wxzxq8xcx"]


def test_witness_v0_witness_v0_scripthash():
    scripthex = "0020bbb988ee4511a6edb5b16f4ea76ed3fd2766b5b60fb6299a9827bc8b9193ad51"
    adress_list, script_type = parse_script(scripthex)
    assert script_type == "p2wshv0"
    assert adress_list == [
        "bc1qhwuc3mj9zxnwmdd3da82wmknl5nkdddkp7mznx5cy77ghyvn44gsnn4qal"
    ]


def test_nulldata():
    scripthex = (
        "6a24b9e11b6d96269d6f48b50c1c17f1ad926aa88fa60629e5ad084df1ef889901eb9cfe11b5"
    )
    adress_list, script_type = parse_script(scripthex)
    assert script_type == "nulldata"
    assert adress_list is None


def test_witness_v1_taproot_is_not_supported():
    scripthex = "5120667bdd93c7c029767fd516d2ea292624b938fefefa175ac9f1220cf508963ff3"
    with pytest.raises(UnknownScriptType):
        _, _ = parse_script(scripthex)


def test_pubkey_but_unparseable():
    # https://www.blockchain.com/btc/tx/71bbaef28e09d8d6fadd41f053db7768dbb5fa4570f06b961dfc29db3dc00b1d
    scripthex = "41617320656d62656464656420696e746f20746865207075626c6963206b65792e000000000000000000000000000000000000000000000000000000000000000000ac"  # noqa
    scripthex = "419e000000416e6f7468657220746578742077617320656d62656464656420696e746f2074686520626c6f636b20636861696e2e20546865207374616e6461726420ac"  # noqa
    with pytest.raises(P2pkParserException):
        _, _ = parse_script(scripthex)


def test_address_conversion():
    for script_type in ["witness_unknown", "nonstandard", "null", "nulldata"]:
        assert address_as_string({"type": script_type}) is None

    for addresses, script_type in [
        (["36TcL12bmahTRrd3dDLRRLQBHYp64xUdng"], "pubkey"),
        (
            [
                "36TcL12bmahTRrd3dDLRRLQBHYp64xUdng",
                "3GCsjdaiebDCBFAoiVv4odtP1dmruRBJCy",
            ],
            "multisig",
        ),
    ]:
        assert (
            address_as_string({"type": script_type, "addresses": addresses})
            == addresses
        )


def test_addresstype_conversion():
    with pytest.raises(UnknownAddressType):
        addresstype_to_int("unknown_type_string")

    assert addresstype_to_int("nonstandard") == 1

    assert addresstype_to_int("pubkey") == 2
    assert addresstype_to_int("p2pk") == 2

    assert addresstype_to_int("pubkeyhash") == 3
    assert addresstype_to_int("p2pkh") == 3

    assert addresstype_to_int("multisig_pubkey") == 4

    assert addresstype_to_int("scripthash") == 5
    assert addresstype_to_int("p2sh") == 5

    assert addresstype_to_int("multisig") == 6

    assert addresstype_to_int("null") == 7
    assert addresstype_to_int("nulldata") == 7

    assert addresstype_to_int("witness_v0_keyhash") == 8
    assert addresstype_to_int("p2wpkhv0") == 8

    assert addresstype_to_int("witness_v0_scripthash") == 9
    assert addresstype_to_int("p2wshv0") == 9

    assert addresstype_to_int("witness_unknown") == 10

    assert addresstype_to_int("witness_v1_taproot") == 11



class UserFacingExceptions(Exception):
    """Hierarchy of exceptions that end up being communicated
    to the end user, but do not produce error logs"""


class NotFoundException(UserFacingExceptions):
    """this exception should be used if some
    item is not found e.g. the database."""


class NetworkNotFoundException(NotFoundException):
    def __init__(self, network):
        super().__init__(f"Network {network} not supported")


class BlockNotFoundException(NotFoundException):
    def __init__(self, network, height):
        super().__init__(f"Block {height} not found in network {network}")


class TransactionNotFoundException(NotFoundException):
    def __init__(self, network, tx_hash, token_id=None):
        msg = (
            (
                f"Token transaction {tx_hash}:{token_id} "
                f"in network {network} not "
                "found"
            )
            if token_id
            else f"Transaction {tx_hash} not found in network {network}"
        )
        super().__init__(msg)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graphsenselib/ingest/tron/grpc/core/contract.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from graphsenselib.ingest.tron.grpc.core import common_pb2 as graphsenselib_dot_ingest_dot_tron_dot_grpc_dot_core_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2graphsenselib/ingest/tron/grpc/core/contract.proto\x12\x08protocol\x1a\x30graphsenselib/ingest/tron/grpc/core/common.proto\"l\n\x15\x41\x63\x63ountCreateContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x17\n\x0f\x61\x63\x63ount_address\x18\x02 \x01(\x0c\x12#\n\x04type\x18\x03 \x01(\x0e\x32\x15.protocol.AccountType\"D\n\x15\x41\x63\x63ountUpdateContract\x12\x14\n\x0c\x61\x63\x63ount_name\x18\x01 \x01(\x0c\x12\x15\n\rowner_address\x18\x02 \x01(\x0c\"A\n\x14SetAccountIdContract\x12\x12\n\naccount_id\x18\x01 \x01(\x0c\x12\x15\n\rowner_address\x18\x02 \x01(\x0c\"\xab\x01\n\x1f\x41\x63\x63ountPermissionUpdateContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12#\n\x05owner\x18\x02 \x01(\x0b\x32\x14.protocol.Permission\x12%\n\x07witness\x18\x03 \x01(\x0b\x32\x14.protocol.Permission\x12%\n\x07\x61\x63tives\x18\x04 \x03(\x0b\x32\x14.protocol.Permission\"\x90\x04\n\x12\x41ssetIssueContract\x12\n\n\x02id\x18) \x01(\t\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\x0c\x12\x0c\n\x04\x61\x62\x62r\x18\x03 \x01(\x0c\x12\x14\n\x0ctotal_supply\x18\x04 \x01(\x03\x12@\n\rfrozen_supply\x18\x05 \x03(\x0b\x32).protocol.AssetIssueContract.FrozenSupply\x12\x0f\n\x07trx_num\x18\x06 \x01(\x05\x12\x11\n\tprecision\x18\x07 \x01(\x05\x12\x0b\n\x03num\x18\x08 \x01(\x05\x12\x12\n\nstart_time\x18\t \x01(\x03\x12\x10\n\x08\x65nd_time\x18\n \x01(\x03\x12\r\n\x05order\x18\x0b \x01(\x03\x12\x12\n\nvote_score\x18\x10 \x01(\x05\x12\x13\n\x0b\x64\x65scription\x18\x14 \x01(\x0c\x12\x0b\n\x03url\x18\x15 \x01(\x0c\x12\x1c\n\x14\x66ree_asset_net_limit\x18\x16 \x01(\x03\x12#\n\x1bpublic_free_asset_net_limit\x18\x17 \x01(\x03\x12#\n\x1bpublic_free_asset_net_usage\x18\x18 \x01(\x03\x12#\n\x1bpublic_latest_free_net_time\x18\x19 \x01(\x03\x1a:\n\x0c\x46rozenSupply\x12\x15\n\rfrozen_amount\x18\x01 \x01(\x03\x12\x13\n\x0b\x66rozen_days\x18\x02 \x01(\x03\"f\n\x15TransferAssetContract\x12\x12\n\nasset_name\x18\x01 \x01(\x0c\x12\x15\n\rowner_address\x18\x02 \x01(\x0c\x12\x12\n\nto_address\x18\x03 \x01(\x0c\x12\x0e\n\x06\x61mount\x18\x04 \x01(\x03\".\n\x15UnfreezeAssetContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\"{\n\x13UpdateAssetContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\x0c\x12\x0b\n\x03url\x18\x03 \x01(\x0c\x12\x11\n\tnew_limit\x18\x04 \x01(\x03\x12\x18\n\x10new_public_limit\x18\x05 \x01(\x03\"n\n\x1dParticipateAssetIssueContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x12\n\nto_address\x18\x02 \x01(\x0c\x12\x12\n\nasset_name\x18\x03 \x01(\x0c\x12\x0e\n\x06\x61mount\x18\x04 \x01(\x03\"\xa3\x01\n\x15\x46reezeBalanceContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x16\n\x0e\x66rozen_balance\x18\x02 \x01(\x03\x12\x17\n\x0f\x66rozen_duration\x18\x03 \x01(\x03\x12(\n\x08resource\x18\n \x01(\x0e\x32\x16.protocol.ResourceCode\x12\x18\n\x10receiver_address\x18\x0f \x01(\x0c\"t\n\x17UnfreezeBalanceContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12(\n\x08resource\x18\n \x01(\x0e\x32\x16.protocol.ResourceCode\x12\x18\n\x10receiver_address\x18\x0f \x01(\x0c\"0\n\x17WithdrawBalanceContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\"M\n\x10TransferContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x12\n\nto_address\x18\x02 \x01(\x0c\x12\x0e\n\x06\x61mount\x18\x03 \x01(\x03\"\xad\x01\n\x17MarketSellAssetContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x15\n\rsell_token_id\x18\x02 \x01(\x0c\x12\x1b\n\x13sell_token_quantity\x18\x03 \x01(\x03\x12\x14\n\x0c\x62uy_token_id\x18\x04 \x01(\x0c\x12\x1a\n\x12\x62uy_token_quantity\x18\x05 \x01(\x03\x12\x15\n\rpre_price_key\x18\x06 \x01(\x0c\"D\n\x19MarketCancelOrderContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x10\n\x08order_id\x18\x02 \x01(\x0c\"^\n\x17ProposalApproveContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x13\n\x0bproposal_id\x18\x02 \x01(\x03\x12\x17\n\x0fis_add_approval\x18\x03 \x01(\x08\"\xa8\x01\n\x16ProposalCreateContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x44\n\nparameters\x18\x02 \x03(\x0b\x32\x30.protocol.ProposalCreateContract.ParametersEntry\x1a\x31\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01\"D\n\x16ProposalDeleteContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x13\n\x0bproposal_id\x18\x02 \x01(\x03\"\x87\x01\n\x13\x43reateSmartContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12-\n\x0cnew_contract\x18\x02 \x01(\x0b\x32\x17.protocol.SmartContract\x12\x18\n\x10\x63\x61ll_token_value\x18\x03 \x01(\x03\x12\x10\n\x08token_id\x18\x04 \x01(\x03\"\x95\x01\n\x14TriggerSmartContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x18\n\x10\x63ontract_address\x18\x02 \x01(\x0c\x12\x12\n\ncall_value\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x18\n\x10\x63\x61ll_token_value\x18\x05 \x01(\x03\x12\x10\n\x08token_id\x18\x06 \x01(\x03\"C\n\x10\x43learABIContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x18\n\x10\x63ontract_address\x18\x02 \x01(\x0c\"o\n\x15UpdateSettingContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x18\n\x10\x63ontract_address\x18\x02 \x01(\x0c\x12%\n\x1d\x63onsume_user_resource_percent\x18\x03 \x01(\x03\"i\n\x19UpdateEnergyLimitContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x18\n\x10\x63ontract_address\x18\x02 \x01(\x0c\x12\x1b\n\x13origin_energy_limit\x18\x03 \x01(\x03\"C\n\x17UpdateBrokerageContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x11\n\tbrokerage\x18\x02 \x01(\x05\"`\n\x11VoteAssetContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x14\n\x0cvote_address\x18\x02 \x03(\x0c\x12\x0f\n\x07support\x18\x03 \x01(\x08\x12\r\n\x05\x63ount\x18\x05 \x01(\x05\";\n\x15WitnessCreateContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x0b\n\x03url\x18\x02 \x01(\x0c\"B\n\x15WitnessUpdateContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x12\n\nupdate_url\x18\x0c \x01(\x0c\"\\\n\x13VoteWitnessContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x1d\n\x05votes\x18\x02 \x03(\x0b\x32\x0e.protocol.Vote\x12\x0f\n\x07support\x18\x03 \x01(\x08\"\x9b\x01\n\x16\x45xchangeCreateContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x16\n\x0e\x66irst_token_id\x18\x02 \x01(\x0c\x12\x1b\n\x13\x66irst_token_balance\x18\x03 \x01(\x03\x12\x17\n\x0fsecond_token_id\x18\x04 \x01(\x0c\x12\x1c\n\x14second_token_balance\x18\x05 \x01(\x03\"e\n\x16\x45xchangeInjectContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x13\n\x0b\x65xchange_id\x18\x02 \x01(\x03\x12\x10\n\x08token_id\x18\x03 \x01(\x0c\x12\r\n\x05quant\x18\x04 \x01(\x03\"g\n\x18\x45xchangeWithdrawContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x13\n\x0b\x65xchange_id\x18\x02 \x01(\x03\x12\x10\n\x08token_id\x18\x03 \x01(\x0c\x12\r\n\x05quant\x18\x04 \x01(\x03\"|\n\x1b\x45xchangeTransactionContract\x12\x15\n\rowner_address\x18\x01 \x01(\x0c\x12\x13\n\x0b\x65xchange_id\x18\x02 \x01(\x03\x12\x10\n\x08token_id\x18\x03 \x01(\x0c\x12\r\n\x05quant\x18\x04 \x01(\x03\x12\x10\n\x08\x65xpected\x18\x05 \x01(\x03\x62\x06proto3')



_ACCOUNTCREATECONTRACT = DESCRIPTOR.message_types_by_name['AccountCreateContract']
_ACCOUNTUPDATECONTRACT = DESCRIPTOR.message_types_by_name['AccountUpdateContract']
_SETACCOUNTIDCONTRACT = DESCRIPTOR.message_types_by_name['SetAccountIdContract']
_ACCOUNTPERMISSIONUPDATECONTRACT = DESCRIPTOR.message_types_by_name['AccountPermissionUpdateContract']
_ASSETISSUECONTRACT = DESCRIPTOR.message_types_by_name['AssetIssueContract']
_ASSETISSUECONTRACT_FROZENSUPPLY = _ASSETISSUECONTRACT.nested_types_by_name['FrozenSupply']
_TRANSFERASSETCONTRACT = DESCRIPTOR.message_types_by_name['TransferAssetContract']
_UNFREEZEASSETCONTRACT = DESCRIPTOR.message_types_by_name['UnfreezeAssetContract']
_UPDATEASSETCONTRACT = DESCRIPTOR.message_types_by_name['UpdateAssetContract']
_PARTICIPATEASSETISSUECONTRACT = DESCRIPTOR.message_types_by_name['ParticipateAssetIssueContract']
_FREEZEBALANCECONTRACT = DESCRIPTOR.message_types_by_name['FreezeBalanceContract']
_UNFREEZEBALANCECONTRACT = DESCRIPTOR.message_types_by_name['UnfreezeBalanceContract']
_WITHDRAWBALANCECONTRACT = DESCRIPTOR.message_types_by_name['WithdrawBalanceContract']
_TRANSFERCONTRACT = DESCRIPTOR.message_types_by_name['TransferContract']
_MARKETSELLASSETCONTRACT = DESCRIPTOR.message_types_by_name['MarketSellAssetContract']
_MARKETCANCELORDERCONTRACT = DESCRIPTOR.message_types_by_name['MarketCancelOrderContract']
_PROPOSALAPPROVECONTRACT = DESCRIPTOR.message_types_by_name['ProposalApproveContract']
_PROPOSALCREATECONTRACT = DESCRIPTOR.message_types_by_name['ProposalCreateContract']
_PROPOSALCREATECONTRACT_PARAMETERSENTRY = _PROPOSALCREATECONTRACT.nested_types_by_name['ParametersEntry']
_PROPOSALDELETECONTRACT = DESCRIPTOR.message_types_by_name['ProposalDeleteContract']
_CREATESMARTCONTRACT = DESCRIPTOR.message_types_by_name['CreateSmartContract']
_TRIGGERSMARTCONTRACT = DESCRIPTOR.message_types_by_name['TriggerSmartContract']
_CLEARABICONTRACT = DESCRIPTOR.message_types_by_name['ClearABIContract']
_UPDATESETTINGCONTRACT = DESCRIPTOR.message_types_by_name['UpdateSettingContract']
_UPDATEENERGYLIMITCONTRACT = DESCRIPTOR.message_types_by_name['UpdateEnergyLimitContract']
_UPDATEBROKERAGECONTRACT = DESCRIPTOR.message_types_by_name['UpdateBrokerageContract']
_VOTEASSETCONTRACT = DESCRIPTOR.message_types_by_name['VoteAssetContract']
_WITNESSCREATECONTRACT = DESCRIPTOR.message_types_by_name['WitnessCreateContract']
_WITNESSUPDATECONTRACT = DESCRIPTOR.message_types_by_name['WitnessUpdateContract']
_VOTEWITNESSCONTRACT = DESCRIPTOR.message_types_by_name['VoteWitnessContract']
_EXCHANGECREATECONTRACT = DESCRIPTOR.message_types_by_name['ExchangeCreateContract']
_EXCHANGEINJECTCONTRACT = DESCRIPTOR.message_types_by_name['ExchangeInjectContract']
_EXCHANGEWITHDRAWCONTRACT = DESCRIPTOR.message_types_by_name['ExchangeWithdrawContract']
_EXCHANGETRANSACTIONCONTRACT = DESCRIPTOR.message_types_by_name['ExchangeTransactionContract']
AccountCreateContract = _reflection.GeneratedProtocolMessageType('AccountCreateContract', (_message.Message,), {
  'DESCRIPTOR' : _ACCOUNTCREATECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.AccountCreateContract)
  })
_sym_db.RegisterMessage(AccountCreateContract)

AccountUpdateContract = _reflection.GeneratedProtocolMessageType('AccountUpdateContract', (_message.Message,), {
  'DESCRIPTOR' : _ACCOUNTUPDATECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.AccountUpdateContract)
  })
_sym_db.RegisterMessage(AccountUpdateContract)

SetAccountIdContract = _reflection.GeneratedProtocolMessageType('SetAccountIdContract', (_message.Message,), {
  'DESCRIPTOR' : _SETACCOUNTIDCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.SetAccountIdContract)
  })
_sym_db.RegisterMessage(SetAccountIdContract)

AccountPermissionUpdateContract = _reflection.GeneratedProtocolMessageType('AccountPermissionUpdateContract', (_message.Message,), {
  'DESCRIPTOR' : _ACCOUNTPERMISSIONUPDATECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.AccountPermissionUpdateContract)
  })
_sym_db.RegisterMessage(AccountPermissionUpdateContract)

AssetIssueContract = _reflection.GeneratedProtocolMessageType('AssetIssueContract', (_message.Message,), {

  'FrozenSupply' : _reflection.GeneratedProtocolMessageType('FrozenSupply', (_message.Message,), {
    'DESCRIPTOR' : _ASSETISSUECONTRACT_FROZENSUPPLY,
    '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
    # @@protoc_insertion_point(class_scope:protocol.AssetIssueContract.FrozenSupply)
    })
  ,
  'DESCRIPTOR' : _ASSETISSUECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.AssetIssueContract)
  })
_sym_db.RegisterMessage(AssetIssueContract)
_sym_db.RegisterMessage(AssetIssueContract.FrozenSupply)

TransferAssetContract = _reflection.GeneratedProtocolMessageType('TransferAssetContract', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFERASSETCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.TransferAssetContract)
  })
_sym_db.RegisterMessage(TransferAssetContract)

UnfreezeAssetContract = _reflection.GeneratedProtocolMessageType('UnfreezeAssetContract', (_message.Message,), {
  'DESCRIPTOR' : _UNFREEZEASSETCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.UnfreezeAssetContract)
  })
_sym_db.RegisterMessage(UnfreezeAssetContract)

UpdateAssetContract = _reflection.GeneratedProtocolMessageType('UpdateAssetContract', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEASSETCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.UpdateAssetContract)
  })
_sym_db.RegisterMessage(UpdateAssetContract)

ParticipateAssetIssueContract = _reflection.GeneratedProtocolMessageType('ParticipateAssetIssueContract', (_message.Message,), {
  'DESCRIPTOR' : _PARTICIPATEASSETISSUECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ParticipateAssetIssueContract)
  })
_sym_db.RegisterMessage(ParticipateAssetIssueContract)

FreezeBalanceContract = _reflection.GeneratedProtocolMessageType('FreezeBalanceContract', (_message.Message,), {
  'DESCRIPTOR' : _FREEZEBALANCECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.FreezeBalanceContract)
  })
_sym_db.RegisterMessage(FreezeBalanceContract)

UnfreezeBalanceContract = _reflection.GeneratedProtocolMessageType('UnfreezeBalanceContract', (_message.Message,), {
  'DESCRIPTOR' : _UNFREEZEBALANCECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.UnfreezeBalanceContract)
  })
_sym_db.RegisterMessage(UnfreezeBalanceContract)

WithdrawBalanceContract = _reflection.GeneratedProtocolMessageType('WithdrawBalanceContract', (_message.Message,), {
  'DESCRIPTOR' : _WITHDRAWBALANCECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.WithdrawBalanceContract)
  })
_sym_db.RegisterMessage(WithdrawBalanceContract)

TransferContract = _reflection.GeneratedProtocolMessageType('TransferContract', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFERCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.TransferContract)
  })
_sym_db.RegisterMessage(TransferContract)

MarketSellAssetContract = _reflection.GeneratedProtocolMessageType('MarketSellAssetContract', (_message.Message,), {
  'DESCRIPTOR' : _MARKETSELLASSETCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.MarketSellAssetContract)
  })
_sym_db.RegisterMessage(MarketSellAssetContract)

MarketCancelOrderContract = _reflection.GeneratedProtocolMessageType('MarketCancelOrderContract', (_message.Message,), {
  'DESCRIPTOR' : _MARKETCANCELORDERCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.MarketCancelOrderContract)
  })
_sym_db.RegisterMessage(MarketCancelOrderContract)

ProposalApproveContract = _reflection.GeneratedProtocolMessageType('ProposalApproveContract', (_message.Message,), {
  'DESCRIPTOR' : _PROPOSALAPPROVECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ProposalApproveContract)
  })
_sym_db.RegisterMessage(ProposalApproveContract)

ProposalCreateContract = _reflection.GeneratedProtocolMessageType('ProposalCreateContract', (_message.Message,), {

  'ParametersEntry' : _reflection.GeneratedProtocolMessageType('ParametersEntry', (_message.Message,), {
    'DESCRIPTOR' : _PROPOSALCREATECONTRACT_PARAMETERSENTRY,
    '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
    # @@protoc_insertion_point(class_scope:protocol.ProposalCreateContract.ParametersEntry)
    })
  ,
  'DESCRIPTOR' : _PROPOSALCREATECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ProposalCreateContract)
  })
_sym_db.RegisterMessage(ProposalCreateContract)
_sym_db.RegisterMessage(ProposalCreateContract.ParametersEntry)

ProposalDeleteContract = _reflection.GeneratedProtocolMessageType('ProposalDeleteContract', (_message.Message,), {
  'DESCRIPTOR' : _PROPOSALDELETECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ProposalDeleteContract)
  })
_sym_db.RegisterMessage(ProposalDeleteContract)

CreateSmartContract = _reflection.GeneratedProtocolMessageType('CreateSmartContract', (_message.Message,), {
  'DESCRIPTOR' : _CREATESMARTCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.CreateSmartContract)
  })
_sym_db.RegisterMessage(CreateSmartContract)

TriggerSmartContract = _reflection.GeneratedProtocolMessageType('TriggerSmartContract', (_message.Message,), {
  'DESCRIPTOR' : _TRIGGERSMARTCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.TriggerSmartContract)
  })
_sym_db.RegisterMessage(TriggerSmartContract)

ClearABIContract = _reflection.GeneratedProtocolMessageType('ClearABIContract', (_message.Message,), {
  'DESCRIPTOR' : _CLEARABICONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ClearABIContract)
  })
_sym_db.RegisterMessage(ClearABIContract)

UpdateSettingContract = _reflection.GeneratedProtocolMessageType('UpdateSettingContract', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESETTINGCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.UpdateSettingContract)
  })
_sym_db.RegisterMessage(UpdateSettingContract)

UpdateEnergyLimitContract = _reflection.GeneratedProtocolMessageType('UpdateEnergyLimitContract', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEENERGYLIMITCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.UpdateEnergyLimitContract)
  })
_sym_db.RegisterMessage(UpdateEnergyLimitContract)

UpdateBrokerageContract = _reflection.GeneratedProtocolMessageType('UpdateBrokerageContract', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEBROKERAGECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.UpdateBrokerageContract)
  })
_sym_db.RegisterMessage(UpdateBrokerageContract)

VoteAssetContract = _reflection.GeneratedProtocolMessageType('VoteAssetContract', (_message.Message,), {
  'DESCRIPTOR' : _VOTEASSETCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.VoteAssetContract)
  })
_sym_db.RegisterMessage(VoteAssetContract)

WitnessCreateContract = _reflection.GeneratedProtocolMessageType('WitnessCreateContract', (_message.Message,), {
  'DESCRIPTOR' : _WITNESSCREATECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.WitnessCreateContract)
  })
_sym_db.RegisterMessage(WitnessCreateContract)

WitnessUpdateContract = _reflection.GeneratedProtocolMessageType('WitnessUpdateContract', (_message.Message,), {
  'DESCRIPTOR' : _WITNESSUPDATECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.WitnessUpdateContract)
  })
_sym_db.RegisterMessage(WitnessUpdateContract)

VoteWitnessContract = _reflection.GeneratedProtocolMessageType('VoteWitnessContract', (_message.Message,), {
  'DESCRIPTOR' : _VOTEWITNESSCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.VoteWitnessContract)
  })
_sym_db.RegisterMessage(VoteWitnessContract)

ExchangeCreateContract = _reflection.GeneratedProtocolMessageType('ExchangeCreateContract', (_message.Message,), {
  'DESCRIPTOR' : _EXCHANGECREATECONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ExchangeCreateContract)
  })
_sym_db.RegisterMessage(ExchangeCreateContract)

ExchangeInjectContract = _reflection.GeneratedProtocolMessageType('ExchangeInjectContract', (_message.Message,), {
  'DESCRIPTOR' : _EXCHANGEINJECTCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ExchangeInjectContract)
  })
_sym_db.RegisterMessage(ExchangeInjectContract)

ExchangeWithdrawContract = _reflection.GeneratedProtocolMessageType('ExchangeWithdrawContract', (_message.Message,), {
  'DESCRIPTOR' : _EXCHANGEWITHDRAWCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ExchangeWithdrawContract)
  })
_sym_db.RegisterMessage(ExchangeWithdrawContract)

ExchangeTransactionContract = _reflection.GeneratedProtocolMessageType('ExchangeTransactionContract', (_message.Message,), {
  'DESCRIPTOR' : _EXCHANGETRANSACTIONCONTRACT,
  '__module__' : 'graphsenselib.ingest.tron.grpc.core.contract_pb2'
  # @@protoc_insertion_point(class_scope:protocol.ExchangeTransactionContract)
  })
_sym_db.RegisterMessage(ExchangeTransactionContract)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PROPOSALCREATECONTRACT_PARAMETERSENTRY._options = None
  _PROPOSALCREATECONTRACT_PARAMETERSENTRY._serialized_options = b'8\001'
  _ACCOUNTCREATECONTRACT._serialized_start=114
  _ACCOUNTCREATECONTRACT._serialized_end=222
  _ACCOUNTUPDATECONTRACT._serialized_start=224
  _ACCOUNTUPDATECONTRACT._serialized_end=292
  _SETACCOUNTIDCONTRACT._serialized_start=294
  _SETACCOUNTIDCONTRACT._serialized_end=359
  _ACCOUNTPERMISSIONUPDATECONTRACT._serialized_start=362
  _ACCOUNTPERMISSIONUPDATECONTRACT._serialized_end=533
  _ASSETISSUECONTRACT._serialized_start=536
  _ASSETISSUECONTRACT._serialized_end=1064
  _ASSETISSUECONTRACT_FROZENSUPPLY._serialized_start=1006
  _ASSETISSUECONTRACT_FROZENSUPPLY._serialized_end=1064
  _TRANSFERASSETCONTRACT._serialized_start=1066
  _TRANSFERASSETCONTRACT._serialized_end=1168
  _UNFREEZEASSETCONTRACT._serialized_start=1170
  _UNFREEZEASSETCONTRACT._serialized_end=1216
  _UPDATEASSETCONTRACT._serialized_start=1218
  _UPDATEASSETCONTRACT._serialized_end=1341
  _PARTICIPATEASSETISSUECONTRACT._serialized_start=1343
  _PARTICIPATEASSETISSUECONTRACT._serialized_end=1453
  _FREEZEBALANCECONTRACT._serialized_start=1456
  _FREEZEBALANCECONTRACT._serialized_end=1619
  _UNFREEZEBALANCECONTRACT._serialized_start=1621
  _UNFREEZEBALANCECONTRACT._serialized_end=1737
  _WITHDRAWBALANCECONTRACT._serialized_start=1739
  _WITHDRAWBALANCECONTRACT._serialized_end=1787
  _TRANSFERCONTRACT._serialized_start=1789
  _TRANSFERCONTRACT._serialized_end=1866
  _MARKETSELLASSETCONTRACT._serialized_start=1869
  _MARKETSELLASSETCONTRACT._serialized_end=2042
  _MARKETCANCELORDERCONTRACT._serialized_start=2044
  _MARKETCANCELORDERCONTRACT._serialized_end=2112
  _PROPOSALAPPROVECONTRACT._serialized_start=2114
  _PROPOSALAPPROVECONTRACT._serialized_end=2208
  _PROPOSALCREATECONTRACT._serialized_start=2211
  _PROPOSALCREATECONTRACT._serialized_end=2379
  _PROPOSALCREATECONTRACT_PARAMETERSENTRY._serialized_start=2330
  _PROPOSALCREATECONTRACT_PARAMETERSENTRY._serialized_end=2379
  _PROPOSALDELETECONTRACT._serialized_start=2381
  _PROPOSALDELETECONTRACT._serialized_end=2449
  _CREATESMARTCONTRACT._serialized_start=2452
  _CREATESMARTCONTRACT._serialized_end=2587
  _TRIGGERSMARTCONTRACT._serialized_start=2590
  _TRIGGERSMARTCONTRACT._serialized_end=2739
  _CLEARABICONTRACT._serialized_start=2741
  _CLEARABICONTRACT._serialized_end=2808
  _UPDATESETTINGCONTRACT._serialized_start=2810
  _UPDATESETTINGCONTRACT._serialized_end=2921
  _UPDATEENERGYLIMITCONTRACT._serialized_start=2923
  _UPDATEENERGYLIMITCONTRACT._serialized_end=3028
  _UPDATEBROKERAGECONTRACT._serialized_start=3030
  _UPDATEBROKERAGECONTRACT._serialized_end=3097
  _VOTEASSETCONTRACT._serialized_start=3099
  _VOTEASSETCONTRACT._serialized_end=3195
  _WITNESSCREATECONTRACT._serialized_start=3197
  _WITNESSCREATECONTRACT._serialized_end=3256
  _WITNESSUPDATECONTRACT._serialized_start=3258
  _WITNESSUPDATECONTRACT._serialized_end=3324
  _VOTEWITNESSCONTRACT._serialized_start=3326
  _VOTEWITNESSCONTRACT._serialized_end=3418
  _EXCHANGECREATECONTRACT._serialized_start=3421
  _EXCHANGECREATECONTRACT._serialized_end=3576
  _EXCHANGEINJECTCONTRACT._serialized_start=3578
  _EXCHANGEINJECTCONTRACT._serialized_end=3679
  _EXCHANGEWITHDRAWCONTRACT._serialized_start=3681
  _EXCHANGEWITHDRAWCONTRACT._serialized_end=3784
  _EXCHANGETRANSACTIONCONTRACT._serialized_start=3786
  _EXCHANGETRANSACTIONCONTRACT._serialized_end=3910
# @@protoc_insertion_point(module_scope)



import logging

from graphsenselib.deltaupdate.update.abstractupdater import (
    TABLE_NAME_DIRTY,
    TABLE_NAME_NEW,
    LegacyUpdateStrategy,
    write_dirty_addresses,
    write_new_addresses,
)

logger = logging.getLogger(__name__)


class UpdateStrategyAccountLegacy(LegacyUpdateStrategy):
    def __init__(self, db, currency, write_new, write_dirty, forward_fill_rates=False):
        super().__init__(
            db, currency, write_new, write_dirty, forward_fill_rates=forward_fill_rates
        )

    def prepare_database(self):
        super().prepare_database()
        ADDRESS_TABLE_COLUMNS = ["address_prefix text", "address blob"]
        ADDRESS_TABLE_PK = ["address_prefix", "address"]

        if self._write_new:
            self._db.transformed.ensure_table_exists(
                TABLE_NAME_NEW,
                ADDRESS_TABLE_COLUMNS
                + ["address_id int", "block_id int", "timestamp int", "tx_hash blob"],
                ADDRESS_TABLE_PK,
                truncate=False,
            )

        if self._write_dirty:
            self._db.transformed.ensure_table_exists(
                TABLE_NAME_DIRTY,
                ADDRESS_TABLE_COLUMNS,
                ADDRESS_TABLE_PK,
                truncate=False,
            )

    def process_batch_impl_hook(self, batch):
        addresses = {}
        bts = self._db.raw.get_block_timestamps_batch(batch)

        for block in batch:
            for row in self._db.raw.get_addresses_in_block(block):
                if row.from_address is not None and row.from_address not in addresses:
                    # From address can be None for coinbase
                    addresses[row.from_address] = row

                if row.to_address is not None and row.to_address not in addresses:
                    # To addresses can be None for e.g. deployments
                    # addresses.add(row.to_address)
                    addresses[row.to_address] = row

        if self._write_dirty:
            write_dirty_addresses(self._db, TABLE_NAME_DIRTY, set(addresses.keys()))

        self._nr_queryied_addresses_batch = len(addresses)
        self._nr_queried_addresses += self._nr_queryied_addresses_batch

        if self._write_new:
            logger.info(
                "Checking existence for "
                f"{self._nr_queryied_addresses_batch} addresses"
            )

            ret = self._db.transformed.known_addresses_batch(
                addresses.keys(), inlcude_new_adresses_table=True
            )

            for addr, exists in ret.items():
                # this is very slow but good for crosschecking
                if not exists:
                    self._new_addresses[addr] = addresses[addr]
                    self._nr_new_addresses += 1

            if self._nr_queried_addresses > 0:
                logger.info(
                    f"Found {self._nr_new_addresses} new addresses so "
                    f"far {(self._nr_new_addresses/self._nr_queried_addresses):.3f} "
                    "are new."
                )
            # logging.info(self._db.transformed.knows_address.cache_info())

            # At the end of the batch write data to the db.
            write_new_addresses(
                self._db,
                TABLE_NAME_NEW,
                [
                    (
                        addr,
                        self.consume_address_id(),
                        int(bts[row.block_id].timestamp()),
                        row.block_id,
                        row.tx_hash,
                    )
                    for addr, row in self._new_addresses.items()
                ],
            )
            # Resetting here means less memory footprint for this job
            # But more writes to Cassandra.
            self._new_addresses = {}



from typing import Iterable, Optional

from ..utils import hex_to_bytes, strip_0x
from .analytics import RawDb, TransformedDb


class TransformedDbAccount(TransformedDb):
    def get_highest_cluster_id(self, sanity_check=True) -> Optional[int]:
        return None

    def get_highest_transaction_id(self):
        return self._get_hightest_id(
            table="transaction_ids_by_transaction_id_group",
            sanity_check=True,
            id_col="transaction_id",
        )


class RawDbAccount(RawDb):
    def get_logs_in_block(self, block: int, topic0=None, contract=None) -> Iterable:
        group = self.get_id_group(block, self.get_block_bucket_size())
        if topic0 is None:
            data = self.select_safe(
                "log", where={"block_id": block, "block_id_group": group}
            )
        else:
            data = self.select_safe(
                "log",
                where={"block_id": block, "block_id_group": group, "topic0": topic0},
            )
        if contract is not None:
            data = [log for log in data if log.address == contract]
        return data

    def get_transaction_ids_in_block(self, block: int) -> Iterable:
        raise NotImplementedError

    def get_transactions_in_block(self, block: int) -> Iterable:
        result = self.select(
            "transaction",
            where={"block_id": block},
        )
        return result

    def get_traces_in_block(self, block: int) -> Iterable:
        block_bucket_size = self.get_block_bucket_size()
        group = self.get_id_group(block, block_bucket_size)

        results = self.select(
            "trace", where={"block_id_group": group, "block_id": block}
        )

        return results

    def get_addresses_in_block(self, block: int) -> Iterable:
        group = self.get_id_group(block, self.get_block_bucket_size())

        # The fetch size is needed since traces currenly contain a lot of null values
        # the null values create tombestones and cassandra refuses to read more than
        # 100k tombestones by default per select, this is avoided by reading in chunks
        # https://community.datastax.com/questions/8110/read-operation-failure-error.html
        # the error happens for example in eth block 15676732 since it has more
        # than 50k traces
        result = self.select(
            "trace",
            columns=["from_address", "to_address", "block_id", "tx_hash"],
            where={"block_id_group": group, "block_id": block},
            fetch_size=10000,
        )
        return result

    def get_tx(self, tx_hash: str) -> object:
        tx_prefix_length = self.get_tx_prefix_length()
        prfix = strip_0x(tx_hash)[:tx_prefix_length]

        result = self.select_one_safe(
            "transaction",
            where={
                "tx_hash_prefix": f"{prfix}",
                "tx_hash": hex_to_bytes(tx_hash),
            },
        )
        return result


class RawDbAccountTrx(RawDbAccount):
    def get_addresses_in_block(self, block: int) -> Iterable:
        group = self.get_id_group(block, self.get_block_bucket_size())

        # The fetch size is needed since traces currenly contain a lot of null values
        # the null values create tombestones and cassandra refuses to read more than
        # 100k tombestones by default per select, this is avoided by reading in chunks
        # https://community.datastax.com/questions/8110/read-operation-failure-error.html
        # the error happens for example in eth block 15676732 since it has more
        # than 50k traces
        result = self.select(
            "trace",
            columns=["caller_address", "transferto_address", "block_id", "tx_hash"],
            where={"block_id_group": group, "block_id": block},
            fetch_size=10000,
        )
        for x in result:
            x["from_address"] = x.pop("caller_address")
            x["to_address"] = x.pop("transferto_address")

        return result
